@@ -118,9 +118,9 @@
   (let ((sxhash-crosscheck
           '#.(let (pairs)
                ;; De-duplicate, which reduces the list from ~8000 entries to ~1000 entries.
-               ;; But make sure that any key which isolated repeated has the same value
+               ;; But make sure that any key which is repeated has the same value
                ;; at each repetition.
-               (dolist (pair sb-c::*sxhash-crosscheck* (coerce pairs 'simple-vector))
+               (dolist (pair *sxhash-crosscheck* (coerce pairs 'simple-vector))
                  (let ((found (assoc (car pair) pairs)))
                    (if found
                        (aver (= (cdr found) (cdr pair)))
@@ -128,7 +128,8 @@
     (loop for (object . hash) across sxhash-crosscheck
           unless (= (sxhash object) hash)
             do (error "SXHASH computed wrong answer for ~S. Got ~x should be ~x"
-                      object hash (sxhash object)))))
+                      object hash (sxhash object))))
+  (format t "~&cross-compiler SXHASH tests passed~%"))
 
 ;;; called when a cold system starts up
 (defun !cold-init ()
@@ -169,14 +170,21 @@
   (/show0 "about to SHOW-AND-CALL !GLOBALDB-COLD-INIT")
   (show-and-call !globaldb-cold-init)
   (show-and-call !function-names-init)
+  (show-and-call !pathname-cold-init)
 
   ;; And now *CURRENT-THREAD*
   (sb-thread::init-main-thread)
 
+  ;; not sure why this is needed on some architectures. Dark magic.
+  (setf (fdefn-fun (find-or-create-fdefn '%coerce-callable-for-call))
+        #'%coerce-callable-to-fun)
+  (show-and-call !loader-cold-init)
   ;; Assert that FBOUNDP doesn't choke when its answer is NIL.
   ;; It was fine if T because in that case the legality of the arg is certain.
   ;; And be extra paranoid - ensure that it really gets called.
-  (locally (declare (notinline fboundp)) (fboundp '(setf !zzzzzz)))
+  (locally
+      (declare (notinline fboundp) (optimize safety)) ; is unsafely flushable
+    (fboundp '(setf !zzzzzz)))
 
   ;; Printing of symbols requires that packages be filled in, because
   ;; OUTPUT-SYMBOL calls FIND-SYMBOL to determine accessibility. Also
@@ -248,6 +256,12 @@
 
   (makunbound '*!cold-toplevels*) ; so it gets GC'd
 
+  ;; Need the static-space replica of the assembly routine jump vector
+  ;; filled in, and the static space vector of static fdefns.
+  ;; This matters only for code that gets compiled to dynamic space,
+  ;; so it's OK that it occurs somewhat late in cold-init.
+  #+x86-64 (sb-vm::validate-asm-routine-vector)
+
   #+win32 (show-and-call reinit-internal-real-time)
 
   ;; Set sane values again, so that the user sees sane values instead
@@ -268,9 +282,8 @@
   (mapcar #'proclaim sb-c::*queued-proclaims*)
   (makunbound 'sb-c::*queued-proclaims*)
 
-  (show-and-call !loader-cold-init)
   (show-and-call os-cold-init-or-reinit)
-  (show-and-call !pathname-cold-init)
+  (show-and-call !lpn-cold-init)
 
   (show-and-call stream-cold-init-or-reset)
   (/show "Enabled buffered streams")
@@ -391,11 +404,13 @@ process to continue normally."
     ;; can be called, as pretty much anything can assume that it is set.
     (when total ; newly started process, and not a failed save attempt
       (sb-thread::init-main-thread)
+      #+x86-64 (sb-vm::validate-asm-routine-vector)
       (rebuild-package-vector))
     ;; Initializing the standard streams calls ALLOC-BUFFER which calls FINALIZE
     (finalizers-reinit)
     ;; Initialize streams next, so that any errors can be printed
     (stream-reinit t)
+    (rebuild-pathname-cache)
     (os-cold-init-or-reinit)
     #-(and win32 (not sb-thread))
     (signal-cold-init-or-reinit)
