@@ -469,6 +469,7 @@ NOTE: This interface is experimental and subject to change."
 ;; Make a new hash-cache and optionally create the statistics vector.
 (defun alloc-hash-cache (size symbol)
   (declare (type index size))
+  (declare (sb-c::tlab :system))
   (let (cache)
     ;; It took me a while to figure out why infinite recursion could occur
     ;; in VALUES-SPECIFIER-TYPE. It's because SET calls VALUES-SPECIFIER-TYPE.
@@ -520,6 +521,7 @@ NOTE: This interface is experimental and subject to change."
                `(progn
                   (declaim (ftype ,ftype ,fn))
                   (defun ,fn ,args
+                    (declare (sb-c::tlab :system))
                     (declare (optimize (safety 0)))
                     ,(if (<= n 3)
                          `(list* ,@args)
@@ -590,7 +592,7 @@ NOTE: This interface is experimental and subject to change."
                              (let* ,(case (length temps)
                                      (2 `((,(first temps) (car ,entry))
                                           (,(second temps) (cdr ,entry))))
-                                     (3 (let ((arg-temp (sb-xc:gensym "ARGS")))
+                                     (3 (let ((arg-temp (gensym "ARGS")))
                                           `((,arg-temp (cdr ,entry))
                                             (,(first temps) (car ,entry))
                                             (,(second temps)
@@ -707,27 +709,18 @@ NOTE: This interface is experimental and subject to change."
                    forms
                    `((,memoizer ,@forms))))))))
 
-;;; FIXME: maybe not the best place
-;;;
-;;; FIXME: think of a better name -- not only does this not have the
-;;; CAR recursion of EQUAL, it also doesn't have the special treatment
-;;; of pathnames, bit-vectors and strings.
-;;;
-;;; KLUDGE: This means that we will no longer cache specifiers of the
-;;; form '(INTEGER (0) 4).  This is probably not a disaster.
-;;;
-;;; A helper function for the type system, which is the main user of
-;;; these caches: we must be more conservative than EQUAL for some of
-;;; our equality tests, because MEMBER and friends refer to EQLity.
-;;; So:
-(defun equal-but-no-car-recursion (x y)
-  (do () (())
-    (cond ((eql x y) (return t))
-          ((and (consp x)
-                (consp y)
-                (eql (pop x) (pop y))))
-          (t
-           (return)))))
+(defun list-elts-eq (x y)
+  (loop (when (eq x y) (return t))
+        (when (or (atom x) (atom y)) (return nil))
+        (unless (eq (car x) (car y)) (return nil))
+        (setq x (cdr x)
+              y (cdr y))))
+(defun list-elements-eql (x y)
+  (loop (when (eq x y) (return t))
+        (when (or (atom x) (atom y)) (return nil))
+        (unless (eql (car x) (car y)) (return nil))
+        (setq x (cdr x)
+              y (cdr y))))
 
 ;;;; various operations on names
 
@@ -827,7 +820,7 @@ NOTE: This interface is experimental and subject to change."
           (let* ((name (first spec))
                  (exp-temp (gensym "ONCE-ONLY")))
             `(let ((,exp-temp ,(second spec))
-                   (,name (sb-xc:gensym ,(symbol-name name))))
+                   (,name (gensym ,(symbol-name name))))
                `(let ((,,name ,,exp-temp))
                   ,,(frob (rest specs) body))))))))
 
@@ -1205,13 +1198,6 @@ NOTE: This interface is experimental and subject to change."
 ;;; - SOCKINT::WIN32-IOCTL                         since 1.2.10 (03/2015)    -> Late: 08/2015
 ;;; - SOCKINT::WIN32-SETSOCKOPT                    since 1.2.10 (03/2015)    -> Late: 08/2015
 ;;; - SOCKINT::WIN32-GETSOCKOPT                    since 1.2.10 (03/2015)    -> Late: 08/2015
-;;;
-;;; - SB-C::MERGE-TAIL-CALLS (policy)              since 1.0.53.74 (11/2011) -> Late: 11/2012
-;;;
-;;; LATE:
-;;; - SB-C::STACK-ALLOCATE-DYNAMIC-EXTENT (policy) since 1.0.19.7            -> Final: anytime
-;;; - SB-C::STACK-ALLOCATE-VECTOR (policy)         since 1.0.19.7            -> Final: anytime
-;;; - SB-C::STACK-ALLOCATE-VALUE-CELLS (policy)    since 1.0.19.7            -> Final: anytime
 
 (defun setup-function-in-final-deprecation
     (software version name replacement-spec)
@@ -1417,15 +1403,6 @@ NOTE: This interface is experimental and subject to change."
              (sorted (stable-sort wrapped comparator :key #'cdr)))
         (map-into sorted #'car sorted))))
 
-;;; Ensure basicness if possible, and simplicity always
-(defun possibly-base-stringize (s)
-  (declare (string s))
-  (cond #+(and sb-unicode (not sb-xc-host))
-        ((and (typep s '(array character (*))) (every #'base-char-p s))
-         (coerce s 'base-string))
-        (t
-         (coerce s 'simple-string))))
-
 ;;; Return T if X is an object that always evaluates to itself. Guard against:
 ;;;  (LET ((S 'XXX)) (SET S 9) (UNINTERN S) (IMPORT S 'KEYWORD) (SYMBOL-VALUE S)) => 9
 ;;; whereby :XXX satisfies KEWORDP and has value 9, or maybe even has no value
