@@ -10,6 +10,8 @@
 
 (in-package "SB-VM")
 
+(declaim (ftype (function (integer) (or symbol (eql 0))) symbol-from-tls-index))
+
 (defvar *current-internal-error-context*)
 
 (defmacro with-pinned-context-code-object
@@ -160,12 +162,11 @@
          ,@body)
        ;; Avoid STATICALLY-LINK-CORE from making it harder to redefine.
        (proclaim '(notinline ,name))
-       (let ((fdefn (find-fdefn ',name))
-             (fun (fdefinition ',variant)))
-         (setf (getf ,(symbolicate '+ cpu-feature '-routines+) fdefn) fun)
-         ;; Redifinition at run-time.
-         (when (eq (%fun-name (fdefn-fun fdefn)) ',variant)
-           (setf (fdefn-fun fdefn) fun))))))
+       (let ((fun #',variant))
+         (setf (getf ,(symbolicate '+ cpu-feature '-routines+) ',name) fun)
+         ;; Redefinition at run-time.
+         (when (eq (%fun-name #',name) ',variant)
+           (setf (%symbol-function ',name) fun))))))
 
 (defvar *previous-cpu-routines* nil)
 
@@ -174,14 +175,60 @@
      ,@(loop for (feature detect) on *cpu-features* by #'cddr
              collect
              `(when ,detect
-                (loop for (fdefn definition) on ,(package-symbolicate "SB-VM" '+ feature '-routines+)
+                (loop for (symbol definition) on ,(package-symbolicate "SB-VM" '+ feature '-routines+)
                       by #'cddr
                       do
-                      (push (cons fdefn (fdefn-fun fdefn))
+                      (push (cons symbol (%symbol-function symbol))
                             *previous-cpu-routines*)
-                      (setf (fdefn-fun fdefn) definition))))))
+                      (setf (%symbol-function symbol) definition))))))
 
 (defun restore-cpu-specific-routines ()
-  (loop for (fdefn . fun) in *previous-cpu-routines*
-        do (setf (fdefn-fun fdefn) fun))
+  (loop for (symbol . fun) in *previous-cpu-routines*
+        do (setf (%symbol-function symbol) fun))
   (setf *previous-cpu-routines* nil))
+
+(defun hexdump (thing &optional (n-words nil wordsp)
+                            ;; pass NIL explicitly if T crashes on you
+                        (decode t))
+  (declare (notinline %code-fun-offset))
+  (multiple-value-bind (obj addr count)
+      (if (typep thing 'word) ; ambiguous in the edge case, but assume it's
+          ;; an address (though you might be trying to dump a bignum's data)
+          (values nil thing (if wordsp n-words 1))
+          (values
+           thing
+           (logandc2 (get-lisp-obj-address thing) lowtag-mask)
+           (if wordsp
+               n-words
+               (if (and (typep thing 'code-component) (plusp (code-n-entries thing)))
+                   ;; Display up through the first fun header
+                   (+ (code-header-words thing)
+                      (ash (%code-fun-offset thing 0) (- word-shift))
+                      simple-fun-insts-offset)
+                   ;; at most 16 words
+                   (min 16 (ash (primitive-object-size thing) (- word-shift)))))))
+    (with-pinned-objects (obj)
+      (dotimes (i count)
+        (let ((word (sap-ref-word (int-sap addr) (ash i word-shift))))
+          (multiple-value-bind (lispobj ok fmt)
+              (cond ((and (typep thing 'code-component)
+                          (< 1 i (code-header-words thing)))
+                     (values (code-header-ref thing i) t))
+                    #+compact-symbol
+                    ((and (typep thing '(and symbol (not null)))
+                          (= i symbol-name-slot))
+                     (values (list (symbol-package-id thing) (symbol-name thing))
+                             t
+                             "{~{~A,~S~}}"))
+                    (decode
+                     (cond #+system-tlabs ; fingers crossed, assume arena pointers are valid
+                           ((and (is-lisp-pointer word) (find-containing-arena word))
+                            (values (%make-lisp-obj word) t))
+                           (t
+                            (make-lisp-obj word nil)))))
+            (let ((*print-lines* 1)
+                  (*print-pretty* t))
+              (format t "~x: ~v,'0x~:[~; = ~@?~]~%"
+                      (+ addr (ash i word-shift))
+                      (* 2 n-word-bytes)
+                      word ok (or fmt "~S") lispobj))))))))
