@@ -236,10 +236,6 @@
 
 ;;; pathname methods
 
-(defun pathname-sxhash (x)
-  (declare (pathname x))
-  (pathname-key-hash x))
-
 (defmethod print-object ((pathname pathname) stream)
   (let ((namestring (handler-case (namestring pathname)
                       (error nil))))
@@ -260,27 +256,11 @@
                   (%pathname-type pathname)
                   (%pathname-version pathname))))))
 
-;;; Use an unsynchronized weak hash table with explicit locking to "try"
-;;; to intern pathnames. "Try" because there are several factors that preclude
-;;; ensuring uniqueness:
-;;;  - pathname is a STRUCTURE-OBJECT, so users might call COPY-STRUCTURE.
-;;;    We can make that fail by inventing a subtype of instance that is not
-;;;    a STRUCTURE-OBJECT. Whatever that metatype is could be useful for
-;;;    PATTERN (parts of the pathname) as well as THREAD, MUTEX, HASH-TABLE
-;;;    and maybe some other things that yield strange semantics if copied.
-;;;  - the load form methods use MAKE-LOAD-FORM-SAVING-SLOTS thereby
-;;;    bypassing the interning operation. That seems totally fixable.
-;;;
-;;; Additionally, it would be nice if this table would act to reduce
-;;; EQUAL to EQ on pathnames by ensuring that we never intern two distinct
-;;; pathnames that are EQUAL.
-;;; That would require performing some canonicalization immediately which may
-;;; or may not pose a problem for fixing https://bugs.launchpad.net/sbcl/+bug/1834266
-;;; which is to say, if hosts can actually be :UNSPECIFIC, then the version
-;;; collapsing that is done here would not be done.  I don't know if it's that simple.
-;;; If it isn't that simple, then the answer is that hash function and comparator
-;;; function used for the *PATHNAMES* table needs to be more fine-grained than
-;;; the hash value that SXHASH returns for a pathname.
+;;; Pathnames are stored in an open-addressing weak hash-set.
+;;; Ideally there would be only one internal representation of any pathname,
+;;; so that EQUAL on pathnames could reduce to EQ.
+;;; I'm not sure that's possible. For the time being, we use a comparator
+;;; that is stricter (I think) than EQUAL.
 ;;;
 ;;; The spec is actually extremely underspecified in regard to the meaning of
 ;;;  "pathnames that are equal should be functionally equivalent."
@@ -292,24 +272,34 @@
 ;;; Also, on case-sensitive-case-preserving filesystems it's not possible
 ;;; to know which pathnames are equivalent without asking the filesystem.
 ;;;
-;;; This table uses %MAKE-HASH-TABLE, not MAKE-HASH-TABLE, because the latter
-;;; always creates a synchronized table if :WEAKNESS is specified.
-;;; But to correctly use the "put-if-absent" operation, the locking must occur
-;;; *around* the get and put operations. It makes no sense to lock the table around
-;;; individual operations, hence the unsynchronized table.
+;;: TODO: consider similarly interning the DEVICE and TYPE parts
+(define-load-time-global *pn-dir-table* nil)
+(define-load-time-global *pn-table* nil)
+(declaim (type robinhood-hashset *pn-dir-table* *pn-table*))
+
+(defmacro compare-pathname-host/dev/dir/name/type (a b)
+  `(and (eq (%pathname-host ,a) (%pathname-host ,b)) ; Interned
+        ;; dir+hash are EQ-comparable thanks to INTERN-PATHNAME
+        (eq (%pathname-dir+hash ,a) (%pathname-dir+hash ,b))
+        ;; the pathname pieces which are strings aren't interned
+        (compare-component (%pathname-device ,a) (%pathname-device ,b))
+        (compare-component (%pathname-name ,a) (%pathname-name ,b))
+        (compare-component (%pathname-type ,a) (%pathname-type ,b))))
+
+(defun pn-table-dir= (entry key)
+  (or (eq (car entry) (car key)) ; quick win if lists are EQ
+      (and (eq (cdr entry) (cdr key)) ; hashes match
+           (compare-component (car entry) (car key)))))
+(defun pn-table-pn= (entry key)
+  (and (compare-pathname-host/dev/dir/name/type entry key)
+       (eql (%pathname-version entry) (%pathname-version key))))
+
+(defun !pathname-cold-init ()
+  (setq *pn-dir-table* (make-hashset 32 #'pn-table-dir= #'cdr
+                                     :synchronized t :weakness t)
+        *pn-table* (make-hashset 32 #'pn-table-pn= #'pathname-sxhash
+                                 :synchronized t :weakness t)))
 
-(define-load-time-global *pathnames*
-    (let ((h (%make-hash-table (logior (pack-ht-flags-weakness +ht-weak-value+)
-                                       (pack-ht-flags-kind 3)
-                                       hash-table-userfun-flag)
-                               'pathname-key=
-                               #'pathname-key=
-                               #'pathname-key-hash
-                               10
-                               default-rehash-size
-                               $1.0)))
-      (install-hash-table-lock h)
-      h))
 ;;; A pathname is logical if the host component is a logical host.
 ;;; This constructor is used to make an instance of the correct type
 ;;; from parsed arguments.
@@ -319,26 +309,87 @@
   ;; but the arguments given in the X3J13 cleanup issue
   ;; PATHNAME-LOGICAL:ADD seem compelling: we should canonicalize the
   ;; case, and uppercase is the ordinary way to do that.
+  (declare (sb-c::tlab :system))
   (flet ((upcase-maybe (x) (typecase x (string (logical-word-or-lose x)) (t x))))
     (when (typep host 'logical-host)
         (setq device :unspecific
               directory (mapcar #'upcase-maybe directory)
               name (upcase-maybe name)
               type (upcase-maybe type))))
-  (let ((table *pathnames*))
-    (declare (inline !allocate-pathname)) ; for DXability
-    (with-system-mutex ((hash-table-%lock table))
-      (let* ((dir+hash (when directory
-                         (ensure-gethash
-                          directory table
-                          (cons directory (pathname-key-hash directory)))))
-             (key (!allocate-pathname host device dir+hash name type version)))
-        (declare (truly-dynamic-extent key))
-        (or (gethash key table)
-            (let ((key (!allocate-pathname host device dir+hash name type version)))
-              (when (typep host 'logical-host)
-                (setf (%instance-wrapper key) #.(find-layout 'logical-pathname)))
-              (setf (gethash key table) key)))))))
+  (dx-let ((dir-key (cons directory (pathname-sxhash directory))))
+    (declare (inline !allocate-pathname)) ; for DX-allocation
+    (flet ((ensure-heap-string (part) ; return any non-string as-is
+             ;; FIXME: what about pattern pieces and (:HOME "user") ?
+             (cond ((or (not (stringp part)) (read-only-space-obj-p part)) part)
+                   ((dynamic-space-obj-p part) (logically-readonlyize part))
+                   ;; dynamic-extent strings and lisp strings in alien memory are acceptable.
+                   ;; Copies must be made in that case, since we're holding on to them.
+                   (t (let ((l (length part)))
+                        (logically-readonlyize
+                         (replace (typecase part
+                                    (base-string (make-string l :element-type 'base-char))
+                                    (t (make-string l)))
+                                  part)))))))
+      (let* ((dir+hash
+              (if directory ; find the interned dir-key
+                  (hashset-insert-if-absent
+                   *pn-dir-table* dir-key
+                   (lambda (dir)
+                     (cons (mapcar #'ensure-heap-string (car dir)) (cdr dir))))))
+             (pn-key (!allocate-pathname host device dir+hash name type version)))
+        (declare (truly-dynamic-extent pn-key))
+        (hashset-insert-if-absent
+         *pn-table* pn-key
+         (lambda (tmp &aux (host (%pathname-host tmp)))
+           (let ((new (!allocate-pathname
+                       host (%pathname-device tmp)
+                       (%pathname-dir+hash tmp)
+                       (ensure-heap-string (%pathname-name tmp))
+                       (ensure-heap-string (%pathname-type tmp))
+                       (%pathname-version tmp))))
+             (when (typep host 'logical-host)
+               (setf (%instance-wrapper new) #.(find-layout 'logical-pathname)))
+             new)))))))
+
+;;; Weak vectors don't work at all once rendered pseudo-static.
+;;; so in order to weaken the pathname cache, the vectors are copied on restart.
+;;; It may not achieve anything for saved pathnames, since the vector elements
+;;; are themselves pseudo-static, but at least newly made ones aren't immortal.
+(defun rebuild-pathname-cache ()
+  (hashset-rehash *pn-dir-table* nil)
+  (hashset-rehash *pn-table* nil))
+
+(defun show-pn-cache (&aux (*print-pretty* nil) (*package* (find-package "CL-USER")))
+  (without-gcing
+   (dolist (symbol '(*pn-dir-table* *pn-table*))
+     (let* ((hashset (symbol-value symbol))
+            (dirs (hss-cells (hashset-storage *pn-dir-table*)))
+            (v (hss-cells (hashset-storage hashset)))
+            (n (hs-cells-capacity v)))
+       (format t "~&~S: size=~D tombstones=~D unused=~D~%" symbol n
+               (count nil v :end n) (count 0 v :end n))
+       (dotimes (i n)
+         (let ((entry (aref v i)))
+           (unless (member entry '(nil 0))
+             (if (eq symbol '*pn-dir-table*)
+                 (format t "~4d ~3d ~x ~16x ~s~%" i
+                         (generation-of entry) (get-lisp-obj-address entry)
+                         (cdr entry) (car entry))
+                 (format t "~4d ~3d ~x ~16x [~A ~S ~A ~S ~S ~S]~%"
+                         i (generation-of entry) (get-lisp-obj-address entry)
+                         (pathname-sxhash entry)
+                         (let ((host (%pathname-host entry)))
+                           (cond ((logical-host-p host)
+                                  (prin1-to-string (logical-host-name host)))
+                                 ((eq host *physical-host*) "phys")
+                                 (t host)))
+                         (%pathname-device entry)
+                         (acond ((%pathname-dir+hash entry)
+                                 (format nil "@~D" (position it dirs)))
+                                (t "-"))
+                         (%pathname-name entry)
+                         (%pathname-type entry)
+                         (%pathname-version entry))))))))))
 
 ;;; Vector of logical host objects, each of which contains its translations.
 ;;; The vector is never mutated- always a new vector is created when adding
@@ -487,11 +538,14 @@
               (string= this that)))
         (pattern
          ;; PATTERN instances should probably become interned objects
-         ;; so that we can use EQ on them. But that currently has the same
-         ;; problem as PATHAME= has - the cache can be bypassed.
+         ;; so that we can use EQ on them.
          (and (pattern-p that)
               (pattern= this that)))
         (cons
+         ;; Even though directory parts are now reliably interned -
+         ;; and so you might be inclined to think that the "full" comparison
+         ;; could be confined to just the interning operation, that's not so,
+         ;; because we also use COMPARE-COMPONENT in ENOUGH-NAMESTRING.
          (and (consp that)
               (compare-component (car this) (car that))
               (compare-component (cdr this) (cdr that))))
@@ -500,50 +554,34 @@
 
 ;;;; pathname functions
 
-(macrolet ((compare-most-components ()
-             `(and (eq (%pathname-host a) (%pathname-host b)) ; Interned
-                   ;; Unless the pathname cache can be made 100% reliable,
-                   ;; strength-reducing EQUAL to EQ is inadmissible here.
-                   ;; To fix that, MAKE-LOAD-FORM methods need not to bypass
-                   ;; INTERN-PATHNAME.
-                   (let ((dir-a (%pathname-dir+hash a))
-                         (dir-b (%pathname-dir+hash b)))
-                     (or (eq dir-a dir-b)
-                         (compare-component (car dir-a) (car dir-b))))
-                   (compare-component (%pathname-device a) (%pathname-device b))
-                   (compare-component (%pathname-name a) (%pathname-name b))
-                   (compare-component (%pathname-type a) (%pathname-type b)))))
-
-;;; PATHNAME-KEY= can receive two different subsets of keys:
-;;; - non-nil LIST is the directory part of a pathname
-;;; - entire PATHNAME
-(defun pathname-key= (a b)
-  (etypecase a
-    (list (and (listp b) (compare-component a b)))
-    (pathname (and (pathnamep b)
-                   (compare-most-components)
-                   (eql (pathname-version a) (pathname-version b))))))
-
 (defun pathname= (a b)
   (declare (type pathname a b))
   (or (eq a b)
-      (and (compare-most-components)
+      ;; I believe that this is actually the same as EQ on pathnames now,
+      ;; but until I prove it, I'm leaving this code in.
+      (and (compare-pathname-host/dev/dir/name/type a b)
            (or (eq (%pathname-host a) *physical-host*)
                (compare-component (pathname-version a)
-                                  (pathname-version b)))))))
+                                  (pathname-version b))))))
 
 (sb-kernel::assign-equalp-impl 'pathname #'pathname=)
 (sb-kernel::assign-equalp-impl 'logical-pathname #'pathname=)
 
-;;; A pathname key is a key to an entry in *PATHNAMES*, either a pathname
-;;; or a pathname-directory.
-(defun pathname-key-hash (x)
+;;; Hash either a PATHNAME or a PATHNAME-DIRECTORY. This is called byt both SXHASH
+;;; and by the interning of pathnames, which uses a multi-step approaching to
+;;; coalescing shared subparts. If an EQUAL directory was used before, we share that.
+;;; Since a directory is stored with its hash precomputed, hashing a PATHNAME as a
+;;; whole entails at most 4 more MIX operations. So using pathnames as keys in
+;;; a hash-table pays a small up-front price for later speed improvement.
+(defun pathname-sxhash (x)
   (flet ((hash-piece (piece)
            (etypecase piece
              (string (sxhash piece)) ; transformed
              (symbol (sxhash piece)) ; transformed
              (pattern (pattern-hash piece))
              ((cons (eql :home) (cons string null))
+              ;; :HOME has two representations- one is just '(:absolute :home ...)
+              ;; and the other '(:absolute (:home "user") ...)
               (sxhash (second piece))))))
     (etypecase x
       (pathname
@@ -940,31 +978,24 @@ a host-structure or string."
 ;;; parse-logical-namestring through to %PARSE-NAMESTRING as a truth
 ;;; value. Yeah, this is probably a KLUDGE - CSR, 2002-04-18
 (defun parseable-logical-namestring-p (namestr start end)
-  (catch 'exit
-    (handler-bind
-        ((namestring-parse-error (lambda (c)
-                                   (declare (ignore c))
-                                   (throw 'exit nil))))
-      (let ((colon (position #\: namestr :start start :end end)))
-        (when colon
-          (let ((potential-host
-                 (logical-word-or-lose (subseq namestr start colon))))
-            ;; depending on the outcome of CSR comp.lang.lisp post
-            ;; "can PARSE-NAMESTRING create logical hosts", we may need
-            ;; to do things with potential-host (create it
-            ;; temporarily, parse the namestring and unintern the
-            ;; logical host potential-host on failure.
-            (declare (ignore potential-host))
-            (let ((result
-                   (handler-bind
-                       ((simple-type-error (lambda (c)
-                                             (declare (ignore c))
-                                             (throw 'exit nil))))
-                     (parse-logical-namestring namestr start end))))
-              ;; if we got this far, we should have an explicit host
-              ;; (first return value of parse-logical-namestring)
-              (aver result)
-              result)))))))
+  (and (parse-potential-logical-host namestr start end)
+       (handler-case
+           (let ((result (parse-logical-namestring namestr start end)))
+             ;; if we got this far, we should have an explicit host
+             ;; (first return value of parse-logical-namestring)
+             (aver result)
+             result)
+         ((or simple-type-error namestring-parse-error) ()
+           nil))))
+
+(defun parse-potential-logical-host (namestr &optional (start 0) end)
+  (handler-case
+    (let ((colon (position #\: namestr :start start :end end)))
+      (when colon
+        (let ((potential-host
+               (logical-word-or-lose (subseq namestr start colon))))
+          (values potential-host colon))))
+    (namestring-parse-error () nil)))
 
 ;;; Handle the case where PARSE-NAMESTRING is actually parsing a
 ;;; namestring. We pick off the :JUNK-ALLOWED case then find a host to
@@ -1197,7 +1228,7 @@ system's syntax for files."
             (let ((host (pathname-host-or-no-namestring pathname)))
               (setf (%pathname-namestring pathname)
                     (logically-readonlyize
-                     (possibly-base-stringize
+                     (possibly-base-stringize-to-heap
                       (funcall (host-unparse host) pathname)))))))))
 
   (defun host-namestring (pathname)
@@ -1738,10 +1769,8 @@ unspecified elements into a completed to-pathname based on the to-wildname."
 ;;; *DEFAULT-PATHNAME-DEFAULTS* before *DEFAULT-PATHNAME-DEFAULTS* is
 ;;; initialized (at which time we can't safely call e.g. #'PATHNAME).
 (defun make-trivial-default-logical-pathname ()
-  (intern-pathname (make-logical-host :name "") :unspecific nil nil nil nil))
-
-(define-load-time-global *logical-pathname-defaults*
-  (make-trivial-default-logical-pathname))
+  (intern-pathname (load-time-value (make-logical-host :name "") t)
+                   :unspecific nil nil nil nil))
 
 (defun logical-namestring-p (x)
   (and (stringp x)
@@ -1763,13 +1792,16 @@ unspecified elements into a completed to-pathname based on the to-wildname."
                       :expected-type 'logical-namestring
                       :format-control "~S is not a valid logical namestring:~%  ~A"
                       :format-arguments (list pathspec problem))))
-        (let ((res (handler-case
-                       (parse-namestring pathspec nil *logical-pathname-defaults*)
-                     (error (e) (oops e)))))
-          (when (eq (%pathname-host res)
-                    (%pathname-host *logical-pathname-defaults*))
-            (oops "no host specified"))
-          res))))
+        (handler-case
+            (if (streamp pathspec)
+                (pathname pathspec)
+                (let ((potential-host (parse-potential-logical-host pathspec)))
+                  (if potential-host
+                      (values (parse-namestring
+                               pathspec (find-logical-host potential-host)))
+                      (error "no host specified"))))
+          (error (e) (oops e))))))
+
 
 ;;;; logical pathname unparsing
 
@@ -1964,7 +1996,7 @@ experimental and subject to change."
                                                   :version :newest))
                 (read lpt))))))
 
-(defun !pathname-cold-init ()
+(defun !lpn-cold-init ()
   (let* ((sys *default-pathname-defaults*)
          (src
           (merge-pathnames
