@@ -9,10 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!IMPL")
-
-(defun time-reinit ()
-  (reinit-internal-real-time))
+(in-package "SB-IMPL")
 
 ;;; Implemented in unix.lisp and win32.lisp.
 (setf (documentation 'get-internal-real-time 'function)
@@ -145,7 +142,7 @@ format."
   (multiple-value-bind (seconds-west daylight)
       (if time-zone
           (values (* time-zone 60 60) nil)
-          (sb!unix::get-timezone (truncate-to-unix-range universal-time)))
+          (sb-unix::get-timezone (truncate-to-unix-range universal-time)))
     (declare (fixnum seconds-west))
     (multiple-value-bind (weeks secs)
         (truncate (+ (- universal-time seconds-west) seconds-offset)
@@ -193,16 +190,15 @@ format."
           (truncate years 100))
        (truncate (+ years 300) 400))))
 
-(define-load-time-global **days-before-month**
+(defconstant-eqx +days-before-month+
   #.(let ((reversed-result nil)
           (sum 0))
       (push nil reversed-result)
       (dolist (days-in-month '(31 28 31 30 31 30 31 31 30 31 30 31))
         (push sum reversed-result)
         (incf sum days-in-month))
-      (coerce (nreverse reversed-result) 'simple-vector)))
-
-(declaim (type (simple-vector 13) **days-before-month**))
+      (coerce (nreverse reversed-result) 'simple-vector))
+  #'equalp)
 
 (defun encode-universal-time (second minute hour date month year
                                      &optional time-zone)
@@ -224,7 +220,7 @@ format."
                    year))
          (days (+ (1- date)
                   (truly-the (mod 335)
-                             (svref **days-before-month** month))
+                             (svref +days-before-month+ month))
                   (if (> month 2)
                       (leap-years-before (1+ year))
                       (leap-years-before year))
@@ -234,12 +230,12 @@ format."
     (if time-zone
         (setf encoded-time (+ second (* (+ minute (* (+ hours time-zone) 60)) 60)))
         (let* ((secwest-guess
-                 (sb!unix::get-timezone
+                 (sb-unix::get-timezone
                   (truncate-to-unix-range (* hours 60 60))))
                (guess (+ second (* 60 (+ minute (* hours 60)))
                          secwest-guess))
                (secwest
-                 (sb!unix::get-timezone
+                 (sb-unix::get-timezone
                   (truncate-to-unix-range guess))))
           (setf encoded-time (+ guess (- secwest secwest-guess)))))
     encoded-time))
@@ -285,7 +281,7 @@ doing so may interfere with results reported by eg. TIME.")
             ;; Round up so we don't mislead by saying 0.0 seconds of non-GC time...
             (- (ceiling total-run-time-us 1000) gc-run-time-ms)
             (if (zerop real-time-ms)
-                100.0
+                $100.0
                 (float (* 100 (/ (round total-run-time-us 1000) real-time-ms))))
             eval-calls
             lambdas-converted
@@ -314,21 +310,21 @@ normally during operations like SLEEP."
 
 ;;; Return all the data that we want TIME to report.
 (defun time-get-sys-info ()
-  (multiple-value-bind (user sys faults) (sb!sys:get-system-info)
+  (multiple-value-bind (user sys faults) (get-system-info)
     (values user sys faults (get-bytes-consed))))
 
 (defun elapsed-cycles (h0 l0 h1 l1)
   (declare (ignorable h0 l0 h1 l1))
-  #!+cycle-counter
+  #+cycle-counter
   (+ (ash (- h1 h0) 32)
      (- l1 l0))
-  #!-cycle-counter
+  #-cycle-counter
   nil)
 (declaim (inline read-cycle-counter))
 (defun read-cycle-counter ()
-  #!+cycle-counter
-  (sb!vm::%read-cycle-counter)
-  #!-cycle-counter
+  #+cycle-counter
+  (sb-vm::%read-cycle-counter)
+  #-cycle-counter
   (values 0 0))
 
 ;;; This is so that we don't have to worry about the vagaries of
@@ -433,12 +429,8 @@ EXPERIMENTAL: Interface subject to change."
         old-page-faults
         new-page-faults
         real-time-overhead
-        run-utime-overhead
-        run-stime-overhead
-        page-faults-overhead
         old-bytes-consed
         new-bytes-consed
-        cons-overhead
         (fun (if (functionp function) function (fdefinition function))))
     (declare (function fun))
     ;; Calculate the overhead...
@@ -452,14 +444,10 @@ EXPERIMENTAL: Interface subject to change."
     (multiple-value-setq
         (new-run-utime new-run-stime new-page-faults new-bytes-consed)
       (time-get-sys-info))
-    (setq run-utime-overhead (- new-run-utime old-run-utime))
-    (setq run-stime-overhead (- new-run-stime old-run-stime))
-    (setq page-faults-overhead (- new-page-faults old-page-faults))
     (setq old-real-time (get-internal-real-time))
     (setq old-real-time (get-internal-real-time))
     (setq new-real-time (get-internal-real-time))
     (setq real-time-overhead (- new-real-time old-real-time))
-    (setq cons-overhead (- new-bytes-consed old-bytes-consed))
     ;; Now get the initial times.
     (multiple-value-setq
         (old-run-utime old-run-stime old-page-faults old-bytes-consed)
@@ -467,9 +455,8 @@ EXPERIMENTAL: Interface subject to change."
     (setq old-real-time (get-internal-real-time))
     (let ((start-gc-internal-run-time *gc-run-time*)
           (*eval-calls* 0)
-          (sb!c::*lambda-conversions* 0)
+          (sb-c::*lambda-conversions* 0)
           (aborted t))
-      (declare (special *eval-calls* sb!c::*lambda-conversions*))
       (multiple-value-bind (h0 l0) (read-cycle-counter)
         (unwind-protect
              (multiple-value-prog1 (apply fun arguments)
@@ -496,10 +483,12 @@ EXPERIMENTAL: Interface subject to change."
                     ;; cycle counting isn't supported everywhere.
                     (when cycles
                       (note :processor-cycles cycles #'zerop))
-                    (note :lambdas-converted sb!c::*lambda-conversions* #'zerop)
+                    (note :lambdas-converted sb-c::*lambda-conversions* #'zerop)
                     (note :eval-calls *eval-calls* #'zerop)
-                    (note :gc-run-time-ms gc-internal-run-time)
+                    (note :gc-run-time-ms (floor gc-internal-run-time
+                                                 (/ internal-time-units-per-second 1000)))
                     (note :system-run-time-us system-run-time)
                     (note :user-run-time-us user-run-time)
-                    (note :real-time-ms real-time))
+                    (note :real-time-ms (floor real-time
+                                               (/ internal-time-units-per-second 1000))))
                   (apply timer plist))))))))))
