@@ -32,7 +32,7 @@
 (defun c-strings->string-list (c-strings)
   (declare (type (alien (* c-string)) c-strings))
   (let ((reversed-result nil))
-    (dotimes (i most-positive-fixnum (error "argh! can't happen"))
+    (dotimes (i most-positive-fixnum)
       (declare (type index i))
       (let ((c-string (deref c-strings i)))
         (if c-string
@@ -67,6 +67,7 @@
                              "sb_select" ; int-syscall
                              "sb_getitimer" ; syscall*
                              "sb_setitimer" ; syscall*
+                             "sb_clock_gettime" ; alien-funcall
                              "sb_utimes") ; posix
                          :test #'string=)
                  (subseq x 3)
@@ -178,11 +179,15 @@ corresponds to NAME, or NIL if there is none."
   #+win32 (sb-win32:unixlike-open path flags :overlapped overlapped)
   #-win32
   (with-restarted-syscall (value errno)
-    (int-syscall ("open" c-string int int)
-                 path
-                 (logior #+largefile o_largefile
-                         flags)
-                 mode)))
+    (locally
+        (declare (optimize (sb-c::float-accuracy 0)))
+      (let ((result (alien-funcall (extern-alien "open" (function int c-string int &optional int))
+                                   path (logior flags
+                                                #+largefile o_largefile)
+                                   mode)))
+        (if (minusp result)
+            (values nil (get-errno))
+            (values result 0))))))
 
 ;;; UNIX-CLOSE accepts a file descriptor and attempts to close the file
 ;;; associated with it.
@@ -653,8 +658,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
      (declare (type (integer 0 #.fd-setsize) ,n))
      ,@body))
 
-;;;; FIXME: Why have both UNIX-SELECT and UNIX-FAST-SELECT?
-
 ;;; Perform the UNIX select(2) system call.
 (declaim (inline unix-fast-select))
 (defun unix-fast-select (num-descriptors
@@ -680,69 +683,7 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
                (note-dangerous-wait "select(2)"))
              (select (int-sap 0)))))))
 
-;;; UNIX-SELECT accepts sets of file descriptors and waits for an event
-;;; to happen on one of them or to time out.
-(declaim (inline num-to-fd-set fd-set-to-num))
-(defun num-to-fd-set (fdset num)
-  (typecase num
-    (fixnum
-     (setf (deref (slot fdset 'fds-bits) 0) num)
-     (loop for index from 1 below (/ fd-setsize
-                                     sb-vm:n-machine-word-bits)
-           do (setf (deref (slot fdset 'fds-bits) index) 0)))
-    (t
-     (loop for index from 0 below (/ fd-setsize
-                                     sb-vm:n-machine-word-bits)
-           do (setf (deref (slot fdset 'fds-bits) index)
-                    (ldb (byte sb-vm:n-machine-word-bits
-                               (* index sb-vm:n-machine-word-bits))
-                         num))))))
-
-(defun fd-set-to-num (nfds fdset)
-  (if (<= nfds sb-vm:n-machine-word-bits)
-      (deref (slot fdset 'fds-bits) 0)
-      (loop for index below (/ fd-setsize
-                               sb-vm:n-machine-word-bits)
-            sum (ash (deref (slot fdset 'fds-bits) index)
-                     (* index sb-vm:n-machine-word-bits)))))
-
-;;; Examine the sets of descriptors passed as arguments to see whether
-;;; they are ready for reading and writing. See the UNIX Programmer's
-;;; Manual for more information.
-(defun unix-select (nfds rdfds wrfds xpfds to-secs &optional (to-usecs 0))
-  (declare (muffle-conditions compiler-note))
-  (declare (type integer nfds)
-           (type unsigned-byte rdfds wrfds xpfds)
-           (type (or (unsigned-byte 31) null) to-secs)
-           (type (unsigned-byte 31) to-usecs)
-           (optimize (speed 3) (safety 0)))
-  (with-fd-setsize (nfds)
-    (with-alien ((tv (struct timeval))
-                 (rdf (struct fd-set))
-                 (wrf (struct fd-set))
-                 (xpf (struct fd-set)))
-      (cond (to-secs
-             (setf (slot tv 'tv-sec) to-secs
-                   (slot tv 'tv-usec) to-usecs))
-            ((not *interrupts-enabled*)
-             (note-dangerous-wait "select(2)")))
-      (num-to-fd-set rdf rdfds)
-      (num-to-fd-set wrf wrfds)
-      (num-to-fd-set xpf xpfds)
-      (macrolet ((frob (lispvar alienvar)
-                   `(if (zerop ,lispvar)
-                        (int-sap 0)
-                        (alien-sap (addr ,alienvar)))))
-        (syscall ("sb_select" int (* (struct fd-set)) (* (struct fd-set))
-                           (* (struct fd-set)) (* (struct timeval)))
-                 (values result
-                         (fd-set-to-num nfds rdf)
-                         (fd-set-to-num nfds wrf)
-                         (fd-set-to-num nfds xpf))
-                 nfds (frob rdfds rdf) (frob wrfds wrf) (frob xpfds xpf)
-                 (if to-secs (alien-sap (addr tv)) (int-sap 0)))))))
-
-;;; Lisp-side implmentations of FD_FOO macros.
+;;; Lisp-side implementations of FD_FOO macros.
 (declaim (inline fd-set fd-clr fd-isset fd-zero))
 (defun fd-set (offset fd-set)
   (multiple-value-bind (word bit) (floor offset
@@ -1049,7 +990,8 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
   (defun clock-gettime (clockid)
     (declare (type (signed-byte 32) clockid))
     (with-alien ((ts (struct timespec)))
-      (alien-funcall (extern-alien "clock_gettime" (function int int (* (struct timespec))))
+      (alien-funcall (extern-alien #.(libc-name-for "sb_clock_gettime")
+                                   (function int int (* (struct timespec))))
                      clockid (addr ts))
       ;; 'seconds' is definitely a fixnum for 64-bit, because most-positive-fixnum
       ;; can express 1E11 years in seconds.
@@ -1088,7 +1030,7 @@ the UNIX epoch (January 1st 1970.)"
               (delta-nsec (the fixnum (- c-nsec (the fixnum (slot base 'tv-nsec))))))
           (the sb-kernel:internal-time
                (+ (the fixnum (* delta-sec internal-time-units-per-second))
-                  (floor delta-nsec nanoseconds-per-internal-time-unit))))
+                  (truncate delta-nsec nanoseconds-per-internal-time-unit))))
 
   ;; There are two optimizations here that actually matter on 32-bit systems:
   ;;  (1) subtract the epoch from seconds and milliseconds separately,
