@@ -109,6 +109,13 @@
 
 (setq sb-c::*track-full-called-fnames* :minimal) ; Change this as desired
 
+;;; Need to get access these sb-sys symbols unqualified from sb-assem
+;;; during make-host-2. Make-host-1 would have already converted its
+;;; code via DEFSETF so it should be insensitive to this substitution.
+(unintern 'sb-assem::sap-ref-16 'sb-assem)
+(unintern 'sb-assem::sap-ref-32 'sb-assem)
+(import '(sb-sys:sap-ref-16 sb-sys:sap-ref-32) 'sb-assem)
+
 (read-undefined-fun-allowlist)
 (defun parallel-make-host-2 (max-jobs)
   (let ((subprocess-count 0)
@@ -204,7 +211,15 @@
             (do-stems-and-flags (stem flags 2)
               (unless (position :not-target flags)
                 (format t "~&[~3D/~3D] ~40A" (incf n) total-files (stem-remap-target stem))
-                (let ((start (get-internal-real-time)))
+                (let ((sb-c::*force-system-tlab*
+                       (or (search "src/pcl" stem)
+                           (search "src/code/arena" stem)
+                           (search "src/code/debug-int" stem)
+                           (search "src/code/early-defmethod" stem)
+                           (search "src/code/format" stem)
+                           (search "src/code/brothertree" stem)
+                           (search "src/code/avltree" stem)))
+                      (start (get-internal-real-time)))
                   (target-compile-stem stem flags)
                   (let ((elapsed (/ (- (get-internal-real-time) start)
                                     internal-time-units-per-second)))
@@ -217,5 +232,4 @@
              (format t "~&~50t ~f~%" total-time))
            (sb-c::dump/restore-interesting-types 'write)))
      (sb-kernel::write-structure-definitions-as-text
-      (sb-cold:stem-object-path "defstructs.lisp-expr"
-                                '(:extra-artifact) :target-compile)))))
+      (sb-cold:find-bootstrap-file "output/defstructs.lisp-expr" t)))))
