@@ -61,11 +61,7 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
   (do-all-symbols (s)
     (let ((dd (sb-kernel:find-defstruct-description s nil)))
       (when (and dd (not (sb-kernel::dd-null-lexenv-p dd)))
-        (let ((pkg (symbol-package (sb-kernel:dd-name dd))))
-          (unless (member (package-name pkg)
-                          '("SB-RBTREE.WORD" "SB-RBTREE.MAP")
-                          :test 'string=)
-            (push (sb-kernel:dd-name dd) result))))))
+        (push (sb-kernel:dd-name dd) result))))
   (assert (equal result '(sb-c::conset))))
 
 ;;; Assert that genesis preserved shadowing symbols.
@@ -84,7 +80,17 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
       ;; in case there is more than one set of floating-point formats.
       (assert (eq (read stream) :default))
       (sb-kernel::with-float-traps-masked (:overflow :divide-by-zero)
-        (let ((*package* (find-package "SB-KERNEL")))
+        (let ((*readtable* (copy-readtable))
+              (*package* (find-package "SB-KERNEL")))
+          (set-dispatch-macro-character
+           #\# #\. (lambda (stream subchar arg)
+                     (declare (ignore subchar arg))
+                     (let ((expr (read stream t nil t)))
+                       (ecase (car expr)
+                         (sb-kernel:make-single-float
+                          (sb-kernel:make-single-float (second expr)))
+                         (sb-kernel:make-double-float
+                          (sb-kernel:make-double-float (second expr) (third expr)))))))
           (dolist (expr (read stream))
             (destructuring-bind (fun args . result) expr
               (let ((result (if (eq (first result) 'sb-kernel::&values)
@@ -164,10 +170,12 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
                     (ecase (if (boundp '*compile-files-p*) *compile-files-p* t)
                      ((t)
                       (let ((sb-c::*source-namestring* fullname)
+                            (sb-c::*force-system-tlab*
+                             (or (search "src/pcl" stem)
+                                 (search "src/code/aprof" stem)))
                             (sb-ext:*derive-function-types*
-                              (if (search "/pcl/" stem)
-                                  :same-file
-                                  t)))
+                              (unless (search "/pcl/" stem)
+                                t)))
                         (ensure-directories-exist output)
                         ;; Like PROCLAIM-TARGET-OPTIMIZATION in 'compile-cold-sbcl'
                         ;; We should probably stash a copy of the POLICY instance from
@@ -237,7 +245,8 @@ sb-kernel::(rplaca (last *handler-clusters*) (car **initial-handler-clusters**))
     (let (list)
       (sb-int:dohash ((name vop) sb-c::*backend-parsed-vops*)
         (declare (ignore vop))
-        (push (cons (gethash name sb-c::*static-vop-usage-counts* 0) name) list))
+        (unless (char= (char (string name) 0) #\!)
+          (push (cons (gethash name sb-c::*static-vop-usage-counts* 0) name) list)))
       (dolist (cell (sort list #'> :key #'car))
         (format output "~7d ~s~%" (car cell) (cdr cell))))))
 
