@@ -27,8 +27,9 @@
                     (default-structure-print lexenv stream depth))))
              (:copier nil)
              (:constructor make-null-lexenv ())
-             (:constructor make-almost-null-lexenv (%policy handled-conditions
-                                                    flushable lambda parent))
+             (:constructor make-almost-null-lexenv
+                           (%policy handled-conditions flushable lambda parent
+                            &optional user-data))
              (:constructor make-package-lock-lexenv
                            (disabled-package-locks %policy
                             &aux (handled-conditions nil)))
@@ -246,6 +247,10 @@
   (result-specs nil :type list)
   type)
 
+(defstruct (lvar-sequence-bounds-annotation
+            (:include lvar-dependent-annotation)
+            (:copier nil)))
+
 (defstruct (lvar-type-annotation
             (:include lvar-annotation)
             (:copier nil))
@@ -1106,6 +1111,8 @@
   ;; True if this functional was created from an inline expansion. This
   ;; is either T, or the GLOBAL-VAR for which it is an expansion.
   (inline-expanded nil)
+  ;; Is it coming from a top-level NAMED-LAMBDA?
+  (top-level-defun-p nil)
   (ignore nil))
 
 (defun pretty-print-functional (functional stream)
@@ -1424,7 +1431,6 @@
   ;; sorted sets of constraints we like to iterate over
   (eql-var-constraints     nil :type (or null (array t 1)))
   (inheritable-constraints nil :type (or null (array t 1)))
-  (private-constraints     nil :type (or null (array t 1)))
   (equality-constraints    nil :type (or null (array t 1)))
 
   source-form)
@@ -1558,7 +1564,8 @@
 ;;; an MV-COMBINATION isn't COMBINATION-P.
 (defstruct (combination (:include basic-combination)
                         (:constructor make-combination (fun))
-                        (:copier nil)))
+                        (:copier nil))
+  (pass-nargs t :type boolean))
 (defprinter (combination :identity t)
   (fun :prin1 (lvar-uses fun))
   (args :prin1 (mapcar (lambda (x)
