@@ -70,9 +70,10 @@
                     (t
                      (delete-1 tn prev setter))))))
              (used-p (tn)
-               (or (tn-reads tn) (tn-writes tn)
-                   (member (tn-kind tn) '(:component :environment))
-                   (not (zerop (sbit aliases (tn-number tn))))))
+               (and (neq (tn-kind tn) :unused)
+                    (or (tn-reads tn) (tn-writes tn)
+                        (member (tn-kind tn) '(:component :environment))
+                        (not (zerop (sbit aliases (tn-number tn)))))))
              (delete-1 (tn prev setter)
                (if prev
                    (setf (tn-next prev) (tn-next tn))
@@ -226,11 +227,13 @@
           (immediate-constant-sc (constant-value constant))
         ;; currently NULL-OFFSET is used only on ARM64
         (if null-offset
-            (setf (leaf-info constant)
-                  (component-live-tn
-                   (make-wired-tn (primitive-type (leaf-type constant))
-                                  immed
-                                  null-offset)))
+            (let ((tn (component-live-tn
+                       (make-wired-tn (primitive-type (leaf-type constant))
+                                      immed
+                                      null-offset
+                                      (specifier-type 'null)))))
+              (setf (tn-leaf tn) constant
+                    (leaf-info constant) tn))
             (let* ((boxed (or (not immed)
                               (boxed-immediate-sc-p immed)))
                    (component (component-info *component-being-compiled*))
@@ -252,7 +255,9 @@
                              #+immobile-space
                              (let ((val (constant-value constant)))
                                (or (and (symbolp val) (not (sb-vm:static-symbol-p val)))
-                                   (typep val 'wrapper)))))
+                                   (typep val 'wrapper))))
+                         #+(or arm64 x86-64)
+                         (not (eql (constant-value constant) $0f0)))
                 (let ((constants (ir2-component-constants component)))
                   (setf (tn-offset res)
                         (vector-push-extend constant constants))))
@@ -356,7 +361,7 @@
 
 (defun reference-tn-refs (refs write-p)
   (when refs
-    (let* ((first (reference-tn (tn-ref-tn refs) write-p) )
+    (let* ((first (reference-tn (tn-ref-tn refs) write-p))
            (prev first))
       (loop for tn-ref = (tn-ref-across refs) then (tn-ref-across tn-ref)
             while tn-ref
