@@ -134,6 +134,9 @@
     (let ((data temp)
           (patch-loc (gen-label))
           (skip-instrumentation (gen-label)))
+      ;; Don't count allocations to the arena
+      (inst cmp :qword (thread-slot-ea thread-arena-slot) 0)
+      (inst jmp :nz skip-instrumentation)
       (inst mov data (thread-slot-ea thread-profile-data-slot thread-temp))
       (inst test data data)
       ;; This instruction is modified to "JMP :z" when profiling is
@@ -146,9 +149,8 @@
       (push patch-loc (sb-assem::asmstream-alloc-points sb-assem:*asmstream*))
       (inst jmp skip-instrumentation)
       (emit-alignment 3 :long-nop)
-      (let ((helper (if (integerp size)
-                        'enable-alloc-counter
-                        'enable-sized-alloc-counter)))
+      (let ((helper
+             (if (integerp size) 'enable-alloc-counter 'enable-sized-alloc-counter)))
         ;; This jump is always encoded as 5 bytes
         (inst call (if (or (not node) ; assembly routine
                            (sb-c::code-immobile-p node))
@@ -165,6 +167,23 @@
              (inst .skip 8 :long-nop)))
       (emit-label skip-instrumentation))))
 
+(defun system-tlab-p (type node)
+  #-system-tlabs (declare (ignore type node))
+  #+system-tlabs
+  (or sb-c::*force-system-tlab*
+      (and (sb-kernel::wrapper-p type)
+           (let ((typename (classoid-name (wrapper-classoid type))))
+             (or (sb-xc:subtypep typename 'ctype)
+                 (eq typename 'sb-thread::avlnode))))
+      (and node
+           (named-let search-env ((env (sb-c::node-lexenv node)))
+             (dolist (data (sb-c::lexenv-user-data env)
+                           (and (sb-c::lexenv-parent env)
+                                (search-env (sb-c::lexenv-parent env))))
+               (when (and (eq (first data) :declare)
+                          (eq (second data) 'sb-c::tlab))
+                 (return (eq (third data) :system))))))))
+
 ;;; An arbitrary marker for the cons primitive-type, not to be confused
 ;;; with the CONS-TYPE in our type-algebraic sense. Mostly just informs
 ;;; the allocator to use cons_tlab.
@@ -182,23 +201,9 @@
 ;;; 1. what to allocate: type, size, lowtag describe the object
 ;;; 2. where to put the result
 ;;; 3. node (for determining immobile-space-p) and a scratch register or two
-(defvar *use-system-tlab* nil)
-(defun system-tlab-p (node)
-  #-system-tlabs (declare (ignore node))
-  #+system-tlabs
-  (or *use-system-tlab*
-      (and node
-           (named-let search-env ((env (sb-c::node-lexenv node)))
-             (dolist (data (sb-c::lexenv-user-data env)
-                           (and (sb-c::lexenv-parent env)
-                                (search-env (sb-c::lexenv-parent env))))
-               (when (and (eq (first data) :declare)
-                          (eq (second data) 'sb-c::tlab))
-                 (return (eq (third data) :system))))))))
-
 (defun allocation (type size lowtag alloc-tn node temp thread-temp
                    &key overflow
-                   &aux (systemp (system-tlab-p node)))
+                   &aux (systemp (system-tlab-p type node)))
   (declare (ignorable thread-temp))
   (flet ((fallback (size)
            ;; Call an allocator trampoline and get the result in the proper register.
@@ -215,7 +220,7 @@
             (if systemp
                 (if (eql type +cons-primtype+) 'sys-list-alloc-tramp 'sys-alloc-tramp)
                 (if (eql type +cons-primtype+) 'list-alloc-tramp 'alloc-tramp))
-            node t)
+            node)
            (inst pop alloc-tn)))
     (let* ((NOT-INLINE (gen-label))
            (DONE (gen-label))
@@ -327,16 +332,30 @@
                                           (lowtag list-pointer-lowtag))
              ;; TODO: gencgc does not need EMIT-GC-STORE-BARRIER here,
              ;; but other other GC strategies might.
-             `(let ((reg
-                     ;; FIXME: single-float gets placed in the boxed header
-                     ;; rather than just doing an immediate store.
-                     (sc-case ,tn
-                      ((control-stack constant)
-                       (move temp ,tn)
-                       temp)
-                      (t
-                       (encode-value-if-immediate ,tn)))))
-                (storew reg ,list ,slot ,lowtag temp))))
+             `(let* (immediate-value
+                     (reg
+                       (sc-case ,tn
+                         (constant
+                          (unless (and (constant-tn-p ,tn)
+                                       (eql prev-constant (tn-value ,tn)))
+                            (setf prev-constant (if (constant-tn-p ,tn)
+                                                    (tn-value ,tn)
+                                                    temp))
+                            (move temp ,tn))
+                          temp)
+                         (immediate
+                          (if (eql prev-constant (setf immediate-value (tn-value ,tn)))
+                              temp
+                              (encode-value-if-immediate ,tn)))
+                         (control-stack
+                          (setf prev-constant temp) ;; a non-eq initial value
+                          (move temp ,tn)
+                          temp)
+                         (t
+                          ,tn))))
+                (when (eq (storew reg ,list ,slot ,lowtag temp)
+                          temp)
+                  (setf prev-constant immediate-value)))))
 
 (define-vop (cons)
   (:args (car :scs (any-reg descriptor-reg constant immediate))
@@ -348,7 +367,8 @@
   (:node-var node)
   (:generator 10
     (let ((stack-allocate-p (node-stack-allocate-p node))
-          (nbytes (* cons-size n-word-bytes)))
+          (nbytes (* cons-size n-word-bytes))
+          (prev-constant temp)) ;; a non-eq initial value
       (unless stack-allocate-p
         (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn))
       (pseudo-atomic (:elide-if stack-allocate-p :thread-tn thread-tn)
@@ -373,13 +393,15 @@
   (:translate acons)
   (:policy :fast-safe)
   (:generator 10
-    (let ((nbytes (* cons-size 2 n-word-bytes)))
+    (let ((nbytes (* cons-size 2 n-word-bytes))
+          (prev-constant temp))
       (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn)
       (pseudo-atomic (:thread-tn thread-tn)
         (allocation +cons-primtype+ nbytes 0 alloc node temp thread-tn)
         (store-slot tail alloc cons-cdr-slot 0)
         (inst lea temp (ea (+ 16 list-pointer-lowtag) alloc))
         (store-slot temp alloc cons-car-slot 0)
+        (setf prev-constant temp)
         (let ((pair temp) (temp alloc)) ; give STORE-SLOT the ALLOC as its TEMP
           (store-slot key pair)
           (store-slot val pair cons-cdr-slot))
@@ -402,7 +424,8 @@
   (:node-var node)
   (:generator 10
     (let ((stack-allocate-p (node-stack-allocate-p node))
-          (nbytes (* cons-size 2 n-word-bytes)))
+          (nbytes (* cons-size 2 n-word-bytes))
+          (prev-constant temp))
       (unless stack-allocate-p
         (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn))
       (pseudo-atomic (:elide-if stack-allocate-p :thread-tn thread-tn)
@@ -429,7 +452,8 @@
   (:generator 0
     (aver (>= cons-cells 3)) ; prevent regressions in ir2tran's vop selection
     (let ((stack-allocate-p (node-stack-allocate-p node))
-          (size (* (pad-data-block cons-size) cons-cells)))
+          (size (* (pad-data-block cons-size) cons-cells))
+          (prev-constant temp))
       (unless stack-allocate-p
         (instrument-alloc +cons-primtype+ size node (list ptr temp) thread-tn))
       (pseudo-atomic (:elide-if stack-allocate-p :thread-tn thread-tn)
@@ -796,7 +820,7 @@
                        (inst push (if (integerp size) (constantize size) size))
                        (inst push (if (sc-is element immediate) (tn-value element) element))
                        (invoke-asm-routine
-                        'call (if (system-tlab-p node) 'sys-make-list 'make-list) node)
+                        'call (if (system-tlab-p 0 node) 'sys-make-list 'make-list) node)
                        (inst pop result)
                        (inst jmp leave-pa)))
          (compute-end)
@@ -887,16 +911,14 @@
 
 ;;;; automatic allocators for primitive objects
 
-;;; FIXME: figure out how not to need this?
 (define-vop (make-funcallable-instance-tramp)
   (:args)
   (:results (result :scs (any-reg)))
   (:vop-var vop)
   (:generator 1
-    (let ((tramp (make-fixup 'funcallable-instance-tramp :assembly-routine)))
-      (if (sb-c::code-immobile-p vop)
-          (inst lea result (ea tramp rip-tn))
-          (inst mov result tramp)))))
+    ;; gets "... is not the name of a defined VOP." if not defined at all
+    #+compact-instance-header (bug "Shouldn't get here")
+    (inst mov result (make-fixup 'funcallable-instance-tramp :assembly-routine))))
 
 (flet
   ((alloc (name words type lowtag stack-allocate-p result
@@ -919,7 +941,7 @@
              (invoke-asm-routine 'call 'alloc-funinstance vop)
              (inst pop result))
             (t
-             (allocation nil bytes (if type 0 lowtag) result node alloc-temp thread-tn)))
+             (allocation type bytes (if type 0 lowtag) result node alloc-temp thread-tn)))
       (let ((header (compute-object-header words type)))
         (cond #+compact-instance-header
               ((and (eq name '%make-structure-instance) stack-allocate-p)
@@ -1002,7 +1024,7 @@
              `(let ((c-fun (make-fixup ,name :foreign)))
                 (inst call (cond ((sb-c::code-immobile-p node) c-fun)
                                  (t (progn (inst mov rax c-fun) rax)))))))
-(define-vop (alloc-immobile-fixedobj)
+(define-vop (!alloc-immobile-fixedobj)
   (:args (size-class :scs (any-reg) :target c-arg1)
          (nwords :scs (any-reg) :target c-arg2)
          (header :scs (any-reg) :target c-arg3))
@@ -1036,7 +1058,7 @@
 (eval-when (:compile-toplevel)
   (aver (evenp symbol-size))) ; assumptions in the code below
 
-(define-vop (fast-alloc-immobile-symbol)
+(define-vop (!fast-alloc-immobile-symbol)
   (:results (result :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg :offset rax-offset) rax)
   (:temporary (:sc unsigned-reg :offset rbx-offset) rbx)
@@ -1044,9 +1066,9 @@
   (:temporary (:sc unsigned-reg) header)
   (:generator 1
     ;; fixedobj_pages linkage entry: 1 PTE per page, 12-byte struct
-    (inst mov rbx (ea (make-fixup "fixedobj_pages" :foreign-dataref)))
+    (inst mov rbx (rip-relative-ea (make-fixup "fixedobj_pages" :foreign-dataref)))
     ;; fixedobj_page_hint: 1 hint per sizeclass. C type = uint32_t
-    (inst mov rax (ea (make-fixup "fixedobj_page_hint" :foreign-dataref)))
+    (inst mov rax (rip-relative-ea (make-fixup "fixedobj_page_hint" :foreign-dataref)))
     (inst mov rbx (ea rbx)) ; get the base of the fixedobj_pages array
     ;; This has to be pseudoatomic as soon as the page hint is loaded.
     ;; Consider the situation where there is exactly one symbol on a page that
@@ -1063,7 +1085,7 @@
        (inst jmp :z FAIL) ; fail if hint page is 0
        (inst lea rbx (ea rbx rax 8))  ; rbx := &fixedobj_pages[hint].free_index
        ;; compute fixedobj_page_address(hint) into RAX
-       (inst mov rcx (ea (make-fixup "FIXEDOBJ_SPACE_START" :foreign-dataref)))
+       (inst mov rcx (rip-relative-ea (make-fixup "FIXEDOBJ_SPACE_START" :foreign-dataref)))
        (inst shl rax (integer-length (1- immobile-card-bytes)))
        (inst add rax (ea rcx))
        ;; load the page's free pointer
