@@ -46,52 +46,6 @@
     (inst movzx '(:byte :dword) result object)
     DONE))
 
-(macrolet ((read-depthoid ()
-             `(ea (- (+ 4 (ash (+ instance-slots-offset
-                                  (get-dsd-index layout sb-kernel::flags))
-                               word-shift))
-                     instance-pointer-lowtag)
-                  layout)))
-  (define-vop ()
-    (:translate layout-depthoid)
-    (:policy :fast-safe)
-    (:args (layout :scs (descriptor-reg)))
-    (:results (res :scs (any-reg)))
-    (:result-types fixnum)
-    (:generator 1 (inst movsx '(:dword :qword) res (read-depthoid))))
-  (define-vop ()
-    (:translate sb-c::layout-depthoid-ge)
-    (:policy :fast-safe)
-    (:args (layout :scs (descriptor-reg)))
-    (:info k)
-    (:arg-types * (:constant (unsigned-byte 16)))
-    (:conditional :ge)
-    (:generator 1 (inst cmp :dword (read-depthoid) (fixnumize k)))))
-
-(define-vop ()
-  (:translate sb-c::%structure-is-a)
-  (:args (x :scs (descriptor-reg)))
-  (:arg-types * (:constant t))
-  (:info test)
-  (:policy :fast-safe)
-  (:conditional :e)
-  (:generator 1
-    (inst cmp :dword
-          (ea (+ (id-bits-offset)
-                 (ash (- (wrapper-depthoid test) 2) 2)
-                 (- instance-pointer-lowtag))
-              x)
-          ;; Small layout-ids can only occur for layouts made in genesis.
-          ;; Therefore if the compile-time value of the ID is small,
-          ;; it is permanently assigned to that type.
-          ;; Otherwise, we allow for the possibility that the compile-time ID
-          ;; is not the same as the load-time ID.
-          ;; I don't think layout-id 0 can get here, but be sure to exclude it.
-          (if (or (typep (layout-id test) '(and (signed-byte 8) (not (eql 0))))
-                  (not (sb-c::producing-fasl-file)))
-              (layout-id test)
-              (make-fixup test :layout-id)))))
-
 #+compact-instance-header
 (progn
 ;; ~17 instructions vs. 35
@@ -127,8 +81,7 @@
       (inst movzx '(:byte :dword) rax object)
       LOAD-FROM-VECTOR
       (inst mov :dword result
-            (ea (make-fixup '**primitive-object-layouts**
-                           :symbol-value
+            (ea (make-fixup '**primitive-object-layouts** :symbol-value
                            (- (ash vector-data-offset word-shift)
                               other-pointer-lowtag))
                 nil rax 8)) ; no base register
@@ -219,27 +172,34 @@
   (:args (x :scs (descriptor-reg)))
   (:info bits)
   (:arg-types t (:constant (unsigned-byte 8)))
-  (:generator 1 (inst mov :byte (ea (- 1 other-pointer-lowtag) x) bits)))
+  (:generator 1
+    (inst mov :byte (ea (- (/ array-flags-position n-word-bytes) other-pointer-lowtag) x)
+      bits)))
 (define-vop ()
   (:translate reset-header-bits)
   (:policy :fast-safe)
   (:args (x :scs (descriptor-reg)))
-  (:arg-types t (:constant (unsigned-byte 8)))
+  (:arg-types t (:constant (unsigned-byte 16)))
   (:info bits)
   (:generator 1
-    (inst and :byte (ea (- 1 other-pointer-lowtag) x) (logandc1 bits #xff))))
-(define-vop (test-header-bit)
-  (:translate test-header-bit)
+    (let ((byte 1))
+      (when (> bits #xff)
+        (setf bits (ash bits -8))
+        (setf byte 2))
+      (inst and :byte (ea (- byte other-pointer-lowtag) x) (logandc1 bits #xff)))))
+(define-vop (test-header-data-bit)
+  (:translate test-header-data-bit)
   (:policy :fast-safe)
   (:args (array :scs (descriptor-reg)))
   (:info mask)
   (:arg-types t (:constant t))
   (:conditional :ne)
   (:generator 1
-    ;; Assert that the mask is in header-data byte index 0
-    ;; which is byte index 1 of the whole header word.
-    (aver (typep mask '(unsigned-byte 8)))
-    (inst test :byte (ea (- 1 other-pointer-lowtag) array) mask)))
+    (let ((byte 1))
+      (when (> mask #xff)
+        (setf mask (ash mask -8))
+        (setf byte 2))
+      (inst test :byte (ea (- byte other-pointer-lowtag) array) mask))))
 
 (define-vop (pointer-hash)
   (:translate pointer-hash)
@@ -328,53 +288,27 @@
                result))))
 
 ;;;; symbol frobbing
-(defun load-symbol-info-vector (result symbol)
+(defun load-symbol-dbinfo (result symbol)
   (loadw result symbol symbol-info-slot other-pointer-lowtag)
   ;; If RES has list-pointer-lowtag, take its CDR. If not, use it as-is.
   ;; This CMOV safely reads from memory when it does not move, because if
-  ;; there is an info-vector in the slot, it has at least one element.
-  ;; Use bit index 3 of the lowtag to distinguish list from vector.
-  ;; A vector will have a 1 in that bit.
+  ;; there is a PACKED-INFO in the slot, it has at least 4 data words in total
+  ;; - the header, at least one info descriptor, and at least one datum.
+  ;; And since 3 is odd, that would be aligned up to 4.
+  ;; Use bit index 2 of the lowtag to distinguish list from instance.
+  ;; An instance will have a 0 in that bit.
   ;; This would compile to almost the same code without a VOP,
   ;; but using a jmp around a mov instead.
-  (aver (= (logior list-pointer-lowtag #b1000) other-pointer-lowtag))
-  (inst test :byte result #b1000)
-  (inst cmov :e result
-        (object-slot-ea result cons-cdr-slot list-pointer-lowtag)))
+  (aver (= (logior instance-pointer-lowtag #b0100) list-pointer-lowtag))
+  (inst test :byte result #b0100)
+  (inst cmov :nz result (object-slot-ea result cons-cdr-slot list-pointer-lowtag)))
 
-(define-vop (symbol-info-vector)
+(define-vop (symbol-dbinfo)
   (:policy :fast-safe)
-  (:translate symbol-info-vector)
+  (:translate symbol-dbinfo)
   (:args (x :scs (descriptor-reg)))
   (:results (res :scs (descriptor-reg)))
-  (:generator 1 (load-symbol-info-vector res x)))
-
-(define-vop (symbol-plist)
-  (:policy :fast-safe)
-  (:translate symbol-plist)
-  (:args (x :scs (descriptor-reg)))
-  (:results (res :scs (descriptor-reg)))
-  (:temporary (:sc unsigned-reg) temp)
-  (:generator 1
-    #-ubsan
-    (progn
-    (loadw res x symbol-info-slot other-pointer-lowtag)
-    ;; Instruction pun: (CAR x) is the same as (VECTOR-LENGTH x)
-    ;; so if the info slot holds a vector, this gets a fixnum- it's not a plist.
-    (loadw res res cons-car-slot list-pointer-lowtag)
-    (inst mov temp nil-value)
-    (inst test :byte res fixnum-tag-mask)
-    (inst cmov :e res temp))
-    ;; This way doesn't assume that CAR and VECTOR-LENGTH are the same memory access.
-    ;; (And it's not even clear that using CMOV is preferable)
-    #+ubsan
-    (let ((out (gen-label)))
-      (loadw temp x symbol-info-slot other-pointer-lowtag)
-      (inst mov res nil-value)
-      (inst test :byte temp #b1000) ; if temp-reg-tn is a vector, return NIL
-      (inst jmp :ne out)
-      (loadw res temp cons-car-slot list-pointer-lowtag)
-      (emit-label out))))
+  (:generator 1 (load-symbol-dbinfo res x)))
 
 ;;;; other miscellaneous VOPs
 
@@ -395,18 +329,20 @@
   (:arg-types (:constant signed-byte))
   (:policy :fast-safe)
   (:generator 1
-    (inst mov sap (thread-slot-ea n))))
+    #-gs-seg (inst mov sap (if (= n thread-this-slot) thread-tn (thread-slot-ea n)))
+    #+gs-seg (inst mov sap (thread-slot-ea n))))
 (define-vop (current-thread-offset-sap)
   (:results (sap :scs (sap-reg)))
   (:result-types system-area-pointer)
   (:translate current-thread-offset-sap)
-  (:args (n :scs (any-reg) :target sap))
+  (:args (index :scs (any-reg) :target sap))
   (:arg-types tagged-num)
   (:policy :fast-safe)
   (:generator 2
-    (inst mov sap
-          (ea thread-base-tn
-              n (ash 1 (- word-shift n-fixnum-tag-bits)))))))
+    (let (#+gs-seg (thread-tn nil))
+      (inst mov sap
+            (ea thread-segment-reg thread-tn
+                index (ash 1 (- word-shift n-fixnum-tag-bits))))))))
 
 (define-vop (halt)
   (:generator 1
@@ -576,13 +512,41 @@ number of CPU cycles elapsed as secondary value. EXPERIMENTAL."
     ;; atomic because the immobile gen# is in the same byte
     (inst and :lock :byte (ea (- 1 other-pointer-lowtag) fdefn) #x7f)))
 
+(define-vop (sb-c::mark-covered)
+ (:info index)
+ (:generator 1
+   ;; Can't convert index to a code-relative index until the boxed header length
+   ;; has been determined.
+   (inst store-coverage-mark index)))
+
 (define-vop ()
-  (:translate update-object-layout)
-  (:args (obj :scs (descriptor-reg)))
-  (:results (res :scs (descriptor-reg)))
+  (:translate sb-lockless::get-next)
   (:policy :fast-safe)
+  (:args (node :scs (descriptor-reg)))
+  (:results (next-tagged :scs (descriptor-reg))
+            (next-bits :scs (descriptor-reg)))
+  (:generator 10
+    ;; Read the first user-data slot and convert to a tagged pointer,
+    ;; also returning the raw value as a secondary result
+    (pseudo-atomic ()
+      (inst mov next-bits (ea (- (ash (+ instance-slots-offset instance-data-start)
+                                      word-shift)
+                                 instance-pointer-lowtag)
+                              node))
+      ;; This can'be be LEA because usually the NEXT-BITS are tagged.
+      ;; An untagged value appears only if in mid-deletion.
+      (inst mov next-tagged next-bits)
+      (inst or :byte next-tagged instance-pointer-lowtag))))
+
+(define-vop (switch-to-arena)
+  (:args (x :scs (descriptor-reg immediate)))
+  (:temporary (:sc unsigned-reg :offset rdi-offset :from (:argument 0)) rdi)
+  (:temporary (:sc unsigned-reg :offset rsi-offset) rsi)
   (:vop-var vop)
+  (:ignore rsi)
   (:generator 1
-    (inst push obj)
-    (invoke-asm-routine 'call 'invalid-layout-trap vop)
-    (inst pop res)))
+    (inst mov rdi (if (sc-is x immediate) (tn-value x) x))
+    ;; We could have the vop declare that all C volatile regs are clobbered,
+    ;; but since this needs pseudo-atomic decoration anyway, it saves code size
+    ;; to make an assembly routine that preserves all registers.
+    (invoke-asm-routine 'call 'switch-to-arena vop)))
