@@ -159,8 +159,7 @@
                          (convert-to-system-type type2))))))))
 
 (defun make-class-symbol (class-name)
-  (format-symbol #.(find-package "SB-PCL")
-                 "*THE-CLASS-~A*" (symbol-name class-name)))
+  (pcl-format-symbol "*THE-CLASS-~A*" (symbol-name class-name)))
 
 (defvar *standard-method-combination*)
 (defvar *or-method-combination*)
@@ -169,6 +168,10 @@
   (getf (object-plist object) name))
 
 (defun (setf plist-value) (new-value object name)
+  ;; HACK: exploit the fact we know the expansion of (SETF (GETF ...)) involves %PUTF.
+  ;; This keeps PCL plists in the system TLAB if this whole file is compiled in a global
+  ;; policy that does so. REMF is ok since it's a non-consing destructive operation.
+  (declare (inline sb-impl::%putf))
   (if new-value
       (setf (getf (object-plist object) name) new-value)
       (progn
@@ -350,8 +353,6 @@
 
 (defclass standard-reader-method (standard-accessor-method) ())
 (defclass standard-writer-method (standard-accessor-method) ())
-;;; an extension, apparently.
-(defclass standard-boundp-method (standard-accessor-method) ())
 
 ;;; for (SLOT-VALUE X 'FOO) / ACCESSOR-SLOT-VALUE optimization, which
 ;;; can't be STANDARD-READER-METHOD because there is no associated
@@ -363,17 +364,24 @@
 (defclass method-combination (metaobject)
   ((%documentation :initform nil :initarg :documentation)))
 
-(defun make-gf-hash-table ()
+;;; TODO: we can express MAKE-GF-HASHSET in terms of the robinhood hashset
+;;; using the same stable FSC-INSTANCE-HASH as this.
+(defun make-gf-hashset ()
+  ;; Return what is logically a weak hashset, but physically a weak hash-table
+  ;; because we don't implement hashsets.
   (make-hash-table :test 'eq
                    :hash-function #'fsc-instance-hash ; stable hash
                    :weakness :key
                    :synchronized t))
+(defun add-to-weak-hashset (key set) (setf (gethash key set) t))
+(defun remove-from-weak-hashset (key set) (remhash key set))
+(defun weak-hashset-memberp (key set) (gethash key set))
 
 (defclass standard-method-combination (definition-source-mixin
                                        method-combination)
   ((type-name :reader method-combination-type-name :initarg :type-name)
    (options :reader method-combination-options :initarg :options)
-   (%generic-functions :initform (make-gf-hash-table)
+   (%generic-functions :initform (make-gf-hashset)
                        :reader method-combination-%generic-functions)))
 
 (defclass long-method-combination (standard-method-combination)
