@@ -52,11 +52,9 @@
                `(defmethod ,name ,args
                  (declare (ignore initargs))
                  (error 'metaobject-initialization-violation
-                  ;; FIXME: I'm pretty sure this wants to be "~~@<~A~~@:>"
-                  :format-control ,(coerce (format nil "~@<~A~@:>" control)
-                                           'base-string)
-                  :format-arguments (list ',name)
-                  :references '((:amop :initialization method))))))
+                        :format-control ,(format nil "~~@<~A~~@:>" control)
+                        :format-arguments (list ',name)
+                        :references '((:amop :initialization method))))))
   (def reinitialize-instance ((method method) &rest initargs)
     "Method objects cannot be redefined by ~S.")
   (def change-class ((method method) new &rest initargs)
@@ -97,6 +95,16 @@
     (invalid-method-initarg method "~@<~S of ~S is not a ~S.~@:>"
                             :function fun 'function)))
 
+(macrolet ((dolist-carefully ((var list improper-list-handler) &body body)
+             `(let ((,var nil)
+                    (.dolist-carefully. ,list))
+                (loop (when (null .dolist-carefully.) (return nil))
+                   (if (consp .dolist-carefully.)
+                       (progn
+                         (setq ,var (pop .dolist-carefully.))
+                         ,@body)
+                       (,improper-list-handler))))))
+
 (defun check-qualifiers (method qualifiers)
   (flet ((improper-list ()
            (invalid-method-initarg method
@@ -137,6 +145,7 @@
                      ~V[~;~1{~S~}~;~1{~S and ~S~}~:;~{~#[~;and ~]~S~^, ~}~] ~
                      as ~2:*~V[~;a specializer~:;specializers~].~@:>"
                     (length frcs) frcs)))))
+) ; end MACROLET
 
 (defmethod shared-initialize :before
     ((method standard-method) slot-names &key
@@ -167,8 +176,6 @@
   (declare (ignore slot-names method-cell))
   (initialize-method-function initargs method))
 
-(define-load-time-global *the-class-generic-function*
-  (find-class 'generic-function))
 (define-load-time-global *the-class-standard-generic-function*
   (find-class 'standard-generic-function))
 
@@ -249,36 +256,37 @@
 
 (defun real-get-method (generic-function qualifiers specializers
                         &optional (errorp t)
-                        always-check-specializers)
-  (let ((specializer-count (length specializers))
-        (methods (generic-function-methods generic-function)))
-    (when (or methods always-check-specializers)
-      (let ((required-parameter-count
-             (length (arg-info-metatypes (gf-arg-info generic-function)))))
-        ;; Since we internally bypass FIND-METHOD by using GET-METHOD
-        ;; instead we need to do this here or users may get hit by a
-        ;; failed AVER instead of a sensible error message.
-        (unless (= specializer-count required-parameter-count)
-          (error
-           'find-method-length-mismatch
-           :format-control   "~@<The generic function ~S takes ~D ~
+                                  always-check-specializers)
+  (sb-thread::with-recursive-system-lock ((gf-lock generic-function))
+    (let ((specializer-count (length specializers))
+          (methods (generic-function-methods generic-function)))
+      (when (or methods always-check-specializers)
+        (let ((required-parameter-count
+                (length (arg-info-metatypes (gf-arg-info generic-function)))))
+          ;; Since we internally bypass FIND-METHOD by using GET-METHOD
+          ;; instead we need to do this here or users may get hit by a
+          ;; failed AVER instead of a sensible error message.
+          (unless (= specializer-count required-parameter-count)
+            (error
+             'find-method-length-mismatch
+             :format-control   "~@<The generic function ~S takes ~D ~
                               required argument~:P; was asked to ~
                               find a method with specializers ~:S~@:>"
-           :format-arguments (list generic-function required-parameter-count
-                                   (unparse-specializers generic-function specializers))))))
-    (flet ((congruentp (other-method)
-             (let ((other-specializers (method-specializers other-method)))
-               (aver (= specializer-count (length other-specializers)))
-               (and (equal qualifiers (safe-method-qualifiers other-method))
-                    (every #'same-specializer-p specializers other-specializers)))))
-      (declare (dynamic-extent #'congruentp))
-      (cond ((find-if #'congruentp methods))
-            ((null errorp) nil)
-            (t
-             (error "~@<There is no method on ~S with ~:[no ~
+             :format-arguments (list generic-function required-parameter-count
+                                     (unparse-specializers generic-function specializers))))))
+      (flet ((congruentp (other-method)
+               (let ((other-specializers (method-specializers other-method)))
+                 (aver (= specializer-count (length other-specializers)))
+                 (and (equal qualifiers (safe-method-qualifiers other-method))
+                      (every #'same-specializer-p specializers other-specializers)))))
+        (declare (dynamic-extent #'congruentp))
+        (cond ((find-if #'congruentp methods))
+              ((null errorp) nil)
+              (t
+               (error "~@<There is no method on ~S with ~:[no ~
                      qualifiers~;~:*qualifiers ~:S~] and specializers ~
                      ~:S.~@:>"
-                    generic-function qualifiers specializers))))))
+                      generic-function qualifiers specializers)))))))
 
 (defmethod find-method ((generic-function standard-generic-function)
                         qualifiers specializers &optional (errorp t))
@@ -363,18 +371,6 @@
             (or restp
                 (and number-of-requireds (/= number-of-requireds requireds)))
             specialized-argument-positions)))
-
-(defun make-discriminating-function-arglist (number-required-arguments restp)
-  (nconc (let ((args nil))
-           (dotimes (i number-required-arguments)
-             (push (format-symbol *package* ;; ! is this right?
-                                  "Discriminating Function Arg ~D"
-                                  i)
-                   args))
-           (nreverse args))
-         (when restp
-               `(&rest ,(format-symbol *package*
-                                       "Discriminating Function &rest Arg")))))
 
 (defmethod generic-function-argument-precedence-order
     ((gf standard-generic-function))
@@ -405,7 +401,7 @@
                       :argument-precedence-order argument-precedence-order)
         (set-arg-info gf))
     (let ((mc (generic-function-method-combination gf)))
-      (setf (gethash gf (method-combination-%generic-functions mc)) t))
+      (add-to-weak-hashset gf (method-combination-%generic-functions mc)))
     (when (arg-info-valid-p (slot-value gf 'arg-info))
       (update-dfun gf))))
 
@@ -415,24 +411,25 @@
   (let* ((old-mc (generic-function-method-combination gf))
          (mc (getf args :method-combination old-mc)))
     (unless (eq mc old-mc)
-      (aver (gethash gf (method-combination-%generic-functions old-mc)))
-      (aver (not (gethash gf (method-combination-%generic-functions mc)))))
+      (aver (weak-hashset-memberp gf (method-combination-%generic-functions old-mc)))
+      (aver (not (weak-hashset-memberp gf (method-combination-%generic-functions mc)))))
     (prog1 (call-next-method)
       (unless (eq mc old-mc)
-        (remhash gf (method-combination-%generic-functions old-mc))
-        (setf (gethash gf (method-combination-%generic-functions mc)) t)
+        (remove-from-weak-hashset gf (method-combination-%generic-functions old-mc))
+        (add-to-weak-hashset gf (method-combination-%generic-functions mc))
         (flush-effective-method-cache gf))
-      (cond
-        ((and lambda-list-p apo-p)
-         (set-arg-info gf
-                       :lambda-list lambda-list
-                       :argument-precedence-order argument-precedence-order))
-        (lambda-list-p (set-arg-info gf :lambda-list lambda-list))
-        (t (set-arg-info gf)))
-      (when (arg-info-valid-p (gf-arg-info gf))
-        (update-dfun gf))
-      (map-dependents gf (lambda (dependent)
-                           (apply #'update-dependent gf dependent args))))))
+      (sb-thread::with-recursive-system-lock ((gf-lock gf))
+        (cond
+          ((and lambda-list-p apo-p)
+           (set-arg-info gf
+                         :lambda-list lambda-list
+                         :argument-precedence-order argument-precedence-order))
+          (lambda-list-p (set-arg-info gf :lambda-list lambda-list))
+          (t (set-arg-info gf)))
+        (when (arg-info-valid-p (gf-arg-info gf))
+          (update-dfun gf))
+        (map-dependents gf (lambda (dependent)
+                             (apply #'update-dependent gf dependent args)))))))
 
 (defun set-methods (gf methods)
   (setf (generic-function-methods gf) nil)
@@ -494,27 +491,36 @@
           (setf (info :function :type name) :generic-function))))))
 
 (defun compute-gf-ftype (name)
-  (let ((gf (and (fboundp name) (fdefinition name))))
-    (if (generic-function-p gf)
-        (let* ((ll (generic-function-lambda-list gf))
-               ;; If the GF has &REST without &KEY then we don't augment
-               ;; the FTYPE with keywords, so as not to complain about keywords
-               ;; which seem not to be accepted.
-               (type (sb-c::ftype-from-lambda-list
-                      (if (and (member '&rest ll) (not (member '&key ll)))
-                          ll
-                          (generic-function-pretty-arglist gf)))))
-          ;; It would be nice if globaldb were transactional,
-          ;; so that either both updates or neither occur.
-          (setf (info :function :type name) type
-                (info :function :where-from name) :defined-method)
-          type)
-        ;; The defaulting expression for (:FUNCTION :TYPE) does not store
-        ;; the default. For :GENERIC-FUNCTION that is not FBOUNDP we also
-        ;; don't, however this branch should never be reached because the
-        ;; info only stores :GENERIC-FUNCTION when methods are loaded.
-        ;; Maybe AVER that it does not happen?
-        (sb-c::ftype-from-fdefn name))))
+  (let ((gf (and (fboundp name) (fdefinition name)))
+        (methods-in-compilation-unit (and (boundp 'sb-c::*methods-in-compilation-unit*)
+                                          sb-c::*methods-in-compilation-unit*
+                                          (gethash name sb-c::*methods-in-compilation-unit*))))
+    (cond ((generic-function-p gf)
+           (let* ((ll (generic-function-lambda-list gf))
+                  ;; If the GF has &REST without &KEY then we don't augment
+                  ;; the FTYPE with keywords, so as not to complain about keywords
+                  ;; which seem not to be accepted.
+                  (type (sb-c::ftype-from-lambda-list
+                         (if (and (member '&rest ll) (not (member '&key ll)))
+                             ll
+                             (generic-function-pretty-arglist gf methods-in-compilation-unit)))))
+
+             ;; It would be nice if globaldb were transactional,
+             ;; so that either both updates or neither occur.
+             (setf (info :function :where-from name) :defined-method
+                   (info :function :type name) type)))
+          (methods-in-compilation-unit
+           (setf (info :function :where-from name) :defined-method
+                 (info :function :type name)
+                 (sb-c::ftype-from-lambda-list
+                  (gf-merge-arglists methods-in-compilation-unit))))
+          (t
+           ;; The defaulting expression for (:FUNCTION :TYPE) does not store
+           ;; the default. For :GENERIC-FUNCTION that is not FBOUNDP we also
+           ;; don't, however this branch should never be reached because the
+           ;; info only stores :GENERIC-FUNCTION when methods are loaded.
+           ;; Maybe AVER that it does not happen?
+           (sb-c::ftype-from-definition name)))))
 
 (defun real-add-method (generic-function method &optional skip-dfun-update-p)
   (flet ((similar-lambda-lists-p (old-method new-lambda-list)
@@ -789,13 +795,6 @@
             (invoke-emf emf args))
           (call-no-applicable-method generic-function args)))))
 
-(defun list-eq (x y)
-  (loop (when (atom x) (return (eq x y)))
-        (when (atom y) (return nil))
-        (unless (eq (car x) (car y)) (return nil))
-        (setq x (cdr x)
-              y (cdr y))))
-
 (define-load-time-global *std-cam-methods* nil)
 
 (defun compute-applicable-methods-emf (generic-function)
@@ -804,7 +803,7 @@
              (cam-methods (compute-applicable-methods-using-types
                            cam (list `(eql ,generic-function) t))))
         (values (get-effective-method-function cam cam-methods)
-                (list-eq cam-methods
+                (list-elts-eq cam-methods
                          (or *std-cam-methods*
                              (setq *std-cam-methods*
                                    (compute-applicable-methods-using-types
@@ -866,8 +865,7 @@
                              ((or (eq class *the-class-standard-writer-method*)
                                   (eq class *the-class-global-writer-method*))
                               'writer)
-                             ((or (eq class *the-class-standard-boundp-method*)
-                                  (eq class *the-class-global-boundp-method*))
+                             ((eq class *the-class-global-boundp-method*)
                               'boundp)))))
             (when (and (gf-info-c-a-m-emf-std-p arg-info)
                        type
@@ -899,7 +897,12 @@
          (types1 `((eql ,class) (class-eq ,class) (eql ,slotd)))
          (types (if (eq type 'writer) `(t ,@types1) types1))
          (methods (compute-applicable-methods-using-types gf types))
-         (std-p (null (cdr methods))))
+         (std-p (and (null (cdr methods))
+                     (if (eq type 'boundp)
+                         (null (cdr (compute-applicable-methods-using-types
+                                     (load-time-value #'slot-makunbound-using-class t)
+                                     types)))
+                         t))))
     (values
      (if std-p
          (get-optimized-std-accessor-method-function class slotd type)
@@ -913,9 +916,9 @@
                                    'get-accessor-method-function)))
                     ,optimized-std-fun)))
                 (wrappers
-                 (let ((wrappers (list (layout-of class)
+                 (let ((wrappers (list (wrapper-of class)
                                        (class-wrapper class)
-                                       (layout-of slotd))))
+                                       (wrapper-of slotd))))
                    (if (eq type 'writer)
                        (cons (class-wrapper *the-class-t*) wrappers)
                        wrappers)))
@@ -1085,21 +1088,22 @@
 (defmacro class-test (arg class)
   (cond
     ((eq class *the-class-t*) t)
-    ((eq class *the-class-slot-object*)
-     `(not (typep (classoid-of ,arg) 'system-classoid)))
     ((eq class *the-class-standard-object*)
      `(or (std-instance-p ,arg) (fsc-instance-p ,arg)))
     ((eq class *the-class-funcallable-standard-object*)
      `(fsc-instance-p ,arg))
+    ;; This is going to be cached (in *fgens*),
+    ;; and structure type tests do not check for invalid layout.
+    ;; Cache the wrapper itself, which is going to be different after
+    ;; redifinition.
+    ((structure-class-p class)
+     `(sb-c::%instance-typep ,arg ,(class-wrapper class)))
     (t
      `(typep ,arg ',(class-name class)))))
 
 (defmacro class-eq-test (arg class)
   `(eq (class-of ,arg) ',class))
 
-(defmacro eql-test (arg object)
-  `(eql ,arg ',object))
-
 (defun dnet-methods-p (form)
   (and (consp form)
        (or (eq (car form) 'methods)
@@ -1269,8 +1273,8 @@
 
 (defun compute-secondary-dispatch-function (generic-function net &optional
                                             method-alist wrappers)
-  (function-funcall (compute-secondary-dispatch-function1 generic-function net)
-                    method-alist wrappers))
+  (funcall (the function (compute-secondary-dispatch-function1 generic-function net))
+           method-alist wrappers))
 
 (defvar *eq-case-table-limit* 15)
 (defvar *case-table-limit* 10)
@@ -1420,17 +1424,17 @@
       (multiple-value-bind (cfunction constants)
           ;; We don't want NAMED-LAMBDA for any expressions handed to FNGEN,
           ;; because name mismatches will render the hashing ineffective.
-          (get-fun1 `(lambda ,arglist
+          (get-fun `(lambda ,arglist
                       (declare (optimize (sb-c::store-closure-debug-pointer 3)))
                       ,@(unless function-p
                           `((declare (ignore .pv. .next-method-call.))))
                       (locally (declare #.*optimize-speed*)
-                               (let ((emf ,net))
-                                 ,(make-emf-call nargs applyp 'emf))))
-                    #'net-test-converter
-                    #'net-code-converter
-                    (lambda (form)
-                      (net-constant-converter form generic-function)))
+                        (let ((emf ,net))
+                          ,(make-emf-call nargs applyp 'emf))))
+                   #'net-test-converter
+                   #'net-code-converter
+                   (lambda (form)
+                     (net-constant-converter form generic-function)))
         (lambda (method-alist wrappers)
           (let* ((alist (list nil))
                  (alist-tail alist))
@@ -1573,13 +1577,12 @@
             ;; COMPUTE-DISCRIMINATING-FUNCTION, then (at least for the
             ;; special cases implemented as of 2006-05-09) any information
             ;; in the cache is misplaced.
-            (aver (null dfun-state)))
-          (typecase dfun-state
-            (null
-             (when (eq gf (load-time-value #'compute-applicable-methods t))
-               (update-all-c-a-m-gf-info gf))
-             (cond
-               ((eq gf (load-time-value #'slot-value-using-class t))
+        (aver (null dfun-state)))
+      (typecase dfun-state
+        (null
+         (when (eq gf (load-time-value #'compute-applicable-methods t))
+           (update-all-c-a-m-gf-info gf))
+         (cond ((eq gf (load-time-value #'slot-value-using-class t))
                 (update-slot-value-gf-info gf 'reader)
                 #'slot-value-using-class-dfun)
                ((eq gf (load-time-value #'(setf slot-value-using-class) t))
@@ -1588,6 +1591,9 @@
                ((eq gf (load-time-value #'slot-boundp-using-class t))
                 (update-slot-value-gf-info gf 'boundp)
                 #'slot-boundp-using-class-dfun)
+               ((and (eq gf (load-time-value #'slot-makunbound-using-class t))
+                     (update-slot-value-gf-info (load-time-value #'slot-boundp-using-class t) 'boundp)
+                     nil))
                ;; KLUDGE: PRINT-OBJECT is not a special-case in the sense
                ;; of having a desperately special discriminating function.
                ;; However, it is important that the machinery for printing
@@ -1627,8 +1633,8 @@
                 (make-final-dfun gf))
                (t
                 (make-initial-dfun gf))))
-            (function dfun-state)
-            (cons (car dfun-state))))))
+        (function dfun-state)
+        (cons (car dfun-state))))))
 
 ;;; in general we need to support SBCL's encapsulation for generic
 ;;; functions: the default implementation of encapsulation changes the
@@ -1661,7 +1667,7 @@
    gf (generic-function-encapsulations gf) (call-next-method)))
 
 (defmethod (setf class-name) (new-value class)
-  (let ((classoid (layout-classoid (class-wrapper class))))
+  (let ((classoid (wrapper-classoid (class-wrapper class))))
     (if (and new-value (symbolp new-value))
         (setf (classoid-name classoid) new-value)
         (setf (classoid-name classoid) nil)))
@@ -1689,10 +1695,16 @@
 ;;; The compiler uses this for type-checking that callers pass acceptable
 ;;; keywords, so don't make this do anything fancy like looking at effective
 ;;; methods without also fixing the compiler.
-(defmethod generic-function-pretty-arglist ((gf standard-generic-function))
+(defmethod generic-function-pretty-arglist ((gf standard-generic-function) &optional methods-in-compilation-unit)
   (let ((gf-lambda-list (generic-function-lambda-list gf))
         (methods (generic-function-methods gf)))
-    (flet ((canonize (k)
+    (flet ((lambda-list (m)
+             (or (and methods-in-compilation-unit
+                      (gethash (cons (method-qualifiers m)
+                                     (unparse-specializers gf (method-specializers m)))
+                               methods-in-compilation-unit))
+                 (method-lambda-list m)))
+           (canonize (k)
              (multiple-value-bind (kw var)
                  (parse-key-arg-spec k)
                (if (and (eql (symbol-package kw) *keyword-package*)
@@ -1701,15 +1713,54 @@
                    (list (list kw var))))))
       (multiple-value-bind (llks required optional rest keys)
           (parse-lambda-list gf-lambda-list :silent t)
-        (collect ((keys (mapcar #'canonize keys)))
-        ;; Possibly extend the keyword parameters of the gf by
-        ;; additional key parameters of its methods:
-          (dolist (m methods
-                     (make-lambda-list llks nil required optional rest (keys)))
-            (binding* (((m.llks nil nil nil m.keys)
-                        (parse-lambda-list (method-lambda-list m) :silent t)))
-              (setq llks (logior llks m.llks))
-              (dolist (k m.keys)
-                (unless (member (parse-key-arg-spec k) (keys)
-                                :key #'parse-key-arg-spec :test #'eq)
-                  (keys (canonize k)))))))))))
+        (if (or (ll-kwds-keyp llks)
+                (ll-kwds-restp llks))
+            (collect ((keys (mapcar #'canonize keys)))
+              ;; Possibly extend the keyword parameters of the gf by
+              ;; additional key parameters of its methods:
+              (flet ((process (lambda-list)
+                       (binding* (((m.llks nil nil nil m.keys)
+                                   (parse-lambda-list lambda-list :silent t)))
+                         (setq llks (logior llks m.llks))
+                         (dolist (k m.keys)
+                           (unless (member (parse-key-arg-spec k) (keys)
+                                           :key #'parse-key-arg-spec :test #'eq)
+                             (keys (canonize k)))))))
+                (dolist (m methods)
+                  (process (lambda-list m))))
+              (make-lambda-list llks nil required optional rest (keys)))
+            (make-lambda-list llks nil required optional))))))
+
+(defun gf-merge-arglists (methods-in-compilation-unit)
+  (flet ((canonize (k)
+           (multiple-value-bind (kw var)
+               (parse-key-arg-spec k)
+             (if (and (eql (symbol-package kw) *keyword-package*)
+                      (string= kw var))
+                 var
+                 (list (list kw var))))))
+    (with-hash-table-iterator (iterator methods-in-compilation-unit)
+      (multiple-value-bind (llks required optional rest keys)
+          (parse-lambda-list (nth-value 2 (iterator)) :silent t)
+        (if (or (ll-kwds-keyp llks)
+                (ll-kwds-restp llks))
+            (collect ((keys (mapcar #'canonize keys)))
+              ;; Possibly extend the keyword parameters of the gf by
+              ;; additional key parameters of its methods:
+              (flet ((process (lambda-list)
+                       (binding* (((m.llks nil nil nil m.keys)
+                                   (parse-lambda-list lambda-list :silent t)))
+                         (setq llks (logior llks m.llks))
+                         (dolist (k m.keys)
+                           (unless (member (parse-key-arg-spec k) (keys)
+                                           :key #'parse-key-arg-spec :test #'eq)
+                             (keys (canonize k)))))))
+
+                (loop
+                 (multiple-value-bind (more key value) (iterator)
+                   (declare (ignore key))
+                   (unless more
+                     (return))
+                   (process value)))
+                (make-lambda-list llks nil required optional rest (keys))))
+            (make-lambda-list llks nil required optional))))))
