@@ -50,7 +50,7 @@
 
 ;;; Used for interning parts of SLOT-NAME-LISTS, as part of
 ;;; PV-TABLE interning -- just to save space.
-(define-load-time-global *slot-name-lists* (make-hash-table :test 'equal))
+(define-load-time-global *slot-name-lists* (make-hashset 64 #'list-elts-eq #'sxhash))
 
 ;;; Used for interning PV-TABLES, keyed by the SLOT-NAME-LISTS
 ;;; used.
@@ -58,19 +58,24 @@
 
 ;;; ...and one lock to rule them. Lock because for certain (rare)
 ;;; cases this lock might be grabbed in the course of method dispatch
-;;; -- and mostly this is already under the *world-lock*
+;;; -- and mostly this is already under the **world-lock**
 (define-load-time-global *pv-lock*
   (sb-thread:make-mutex :name "pv table index lock"))
 
 (defun intern-pv-table (&key slot-name-lists)
   (flet ((intern-slot-names (slot-names)
-           (ensure-gethash slot-names *slot-name-lists* slot-names))
+           ;; Hashsets don't like NIL as a key because NIL is the GC-splatted marker
+           ;; (this isn't a weak hashset, but still, don't try to store NIL because
+           ;; it causes the set to size up on every alleged failure to find)
+           (if slot-names
+               (or (hashset-find *slot-name-lists* slot-names)
+                   (hashset-insert *slot-name-lists* slot-names))))
          (%intern-pv-table (snl)
            (ensure-gethash
             snl *pv-tables*
             (make-pv-table :slot-name-lists snl
                            :pv-size (* 2 (reduce #'+ snl :key #'length))))))
-    (sb-thread:with-mutex (*pv-lock*)
+    (with-system-mutex (*pv-lock*)
       (%intern-pv-table (mapcar #'intern-slot-names slot-name-lists)))))
 
 (defun use-standard-slot-access-p (class slot-name type)
@@ -97,7 +102,7 @@
         (when slot-names
           (let* ((wrapper (pop wrappers))
                  (std-p (layout-for-pcl-obj-p wrapper))
-                 (class (wrapper-class* wrapper)))
+                 (class (wrapper-class wrapper)))
             (dolist (slot-name slot-names)
               (destructuring-bind (location . info)
                   (or (find-slot-cell wrapper slot-name)
@@ -312,8 +317,6 @@
 
 (define-walker-template pv-offset) ; These forms get munged by mutate slots.
 (defmacro pv-offset (arg) arg)
-(define-walker-template instance-accessor-parameter)
-(defmacro instance-accessor-parameter (x) x)
 
 ;;; It is safe for these two functions to be wrong. They just try to
 ;;; guess what the most likely case will be.
@@ -384,23 +387,11 @@
        (let ((,index (svref ,pv ,pv-offset))
              (,slots (truly-the simple-vector ,slots)))
          (setq ,value (typecase ,index
-                        ;; FIXME: the line marked by KLUDGE below (and
-                        ;; the analogous spot in
-                        ;; INSTANCE-WRITE-STANDARD) is there purely to
-                        ;; suppress a type mismatch warning that
-                        ;; propagates through to user code.
-                        ;; Presumably SLOTS at this point can never
-                        ;; actually be NIL, but the compiler seems to
-                        ;; think it could, so we put this here to shut
-                        ;; it up.  (see also mail Rudi Schlatte
-                        ;; sbcl-devel 2003-09-21) -- CSR, 2003-11-30
                         ,@(when (or (null kind) (eq kind :instance))
-                                `((fixnum
-                                   (clos-slots-ref ,slots ,index))))
+                            `((fixnum (clos-slots-ref ,slots ,index))))
                         ,@(when (or (null kind) (eq kind :class))
-                                `((cons (cdr ,index))))
-                        (t
-                         +slot-unbound+)))
+                            `((cons (cdr ,index))))
+                        (t +slot-unbound+)))
          (if (unbound-marker-p ,value)
              ,default
              ,value)))))
@@ -450,14 +441,13 @@
               new-value)))
     `(locally (declare #.*optimize-speed*)
        (let ((.good-new-value. ,new-value-form)
-             (,index (svref ,pv ,pv-offset)))
+             (,index (svref ,pv ,pv-offset))
+             (,slots (truly-the simple-vector ,slots)))
          (typecase ,index
            ,@(when (or (null kind) (eq kind :instance))
-                   `((fixnum (and ,slots
-                                  (setf (clos-slots-ref ,slots ,index)
-                                        .good-new-value.)))))
+               `((fixnum (setf (clos-slots-ref ,slots ,index) .good-new-value.))))
            ,@(when (or (null kind) (eq kind :class))
-                   `((cons (setf (cdr ,index) .good-new-value.))))
+               `((cons (setf (cdr ,index) .good-new-value.))))
            (t ,default))))))
 
 (defmacro instance-write-custom (pv pv-offset parameter new-value)
@@ -487,14 +477,13 @@
     (error "illegal kind argument to ~S: ~S" 'instance-boundp-standard kind))
   (let* ((index (gensym)))
     `(locally (declare #.*optimize-speed*)
-       (let ((,index (svref ,pv ,pv-offset)))
+       (let ((,index (svref ,pv ,pv-offset))
+             (,slots (truly-the simple-vector ,slots)))
          (typecase ,index
            ,@(when (or (null kind) (eq kind :instance))
-                   `((fixnum (not (and ,slots
-                                       (unbound-marker-p
-                                        (clos-slots-ref ,slots ,index)))))))
+               `((fixnum (not (unbound-marker-p (clos-slots-ref ,slots ,index))))))
            ,@(when (or (null kind) (eq kind :class))
-                   `((cons (not (unbound-marker-p (cdr ,index))))))
+               `((cons (not (unbound-marker-p (cdr ,index))))))
            (t ,default))))))
 
 (defmacro instance-boundp-custom (pv pv-offset parameter)
@@ -767,8 +756,8 @@
                             ,@inner-decls
                             ,@body-sans-decls))))
                  (mf (%make-method-function fmf)))
-            (set-funcallable-instance-function
-             mf (method-function-from-fast-function fmf ',(getf initargs 'plist)))
+            (setf (%funcallable-instance-fun mf)
+                  (method-function-from-fast-function fmf ',(getf initargs 'plist)))
             mf)
           ',initargs)))))
 
