@@ -548,7 +548,9 @@ check_deferrables_blocked_or_lose(sigset_t *sigset)
         lose("deferrables unblocked");
 }
 
+#ifdef LISP_FEATURE_RISCV
 int sigaction_does_not_mask;
+#endif
 static void assert_blockables_blocked()
 {
 #ifdef LISP_FEATURE_RISCV
@@ -765,7 +767,7 @@ check_interrupt_context_or_lose(os_context_t *context)
 #if defined(LISP_FEATURE_GENCGC) && !GENCGC_IS_PRECISE
     int interrupts_enabled = (read_TLS(INTERRUPTS_ENABLED,thread) != NIL);
     int gc_inhibit = (read_TLS(GC_INHIBIT,thread) != NIL);
-    int gc_pending = (read_TLS(GC_PENDING,thread) == T);
+    int gc_pending = (read_TLS(GC_PENDING,thread) == LISP_T);
     int pseudo_atomic_interrupted = get_pseudo_atomic_interrupted(thread);
     int in_race_p = in_leaving_without_gcing_race_p(thread);
     int safepoint_active = 0;
@@ -1029,15 +1031,13 @@ interrupt_internal_error(os_context_t *context, boolean continuable)
     }
 #endif
 
-    SHOW("in interrupt_internal_error");
-#if QSHOW == 2
     /* Display some rudimentary debugging information about the
      * error, so that even if the Lisp error handler gets badly
      * confused, we have a chance to determine what's going on. */
-    describe_internal_error(context);
-#endif
+    // describe_internal_error(context); // uncomment me for debugging
+
     funcall2(StaticSymbolFunction(INTERNAL_ERROR), context_sap,
-             continuable ? T : NIL);
+             continuable ? LISP_T : NIL);
 
     undo_fake_foreign_function_call(context); /* blocks signals again */
     if (continuable)
@@ -1135,7 +1135,7 @@ interrupt_handle_pending(os_context_t *context)
          /* Test for T and not for != NIL since the value :IN-PROGRESS
           * used to be used in SUB-GC as part of the mechanism to
           * supress recursive gcs.*/
-        if (read_TLS(GC_PENDING,thread) == T) {
+        if (read_TLS(GC_PENDING,thread) == LISP_T) {
 
             /* Two reasons for doing this. First, if there is a
              * pending handler we don't want to run. Second, we are
@@ -1211,7 +1211,7 @@ interrupt_handle_pending(os_context_t *context)
         run_deferred_handler(data, context);
     }
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-    if (read_TLS(THRUPTION_PENDING,thread)==T)
+    if (read_TLS(THRUPTION_PENDING,thread)==LISP_T)
         /* Special case for the following situation: There is a
          * thruption pending, but a signal had been deferred.  The
          * pitstop at the top of this function could only take care
@@ -1326,6 +1326,25 @@ store_signal_data_for_later (struct interrupt_data *data, void *handler,
     sigaddset_deferrable(os_context_sigmask_addr(context));
 }
 
+/* What's going on ?
+ *
+ *  0: fp=0x7fa86423e7a0 pc=0x27608c Foreign function (null)
+ *  1: fp=0x7fa86423e7b0 pc=0x2767aa Foreign function (null)
+ *  2: fp=0x7fa86423e890 pc=0x2761af Foreign function (null)
+ *  3: fp=0x7fa86423e970 pc=0x27882f Foreign function (null)       ; maybe_now_maybe_later
+ *  4: fp=0x7fa86423f5c0 pc=0x7fa8646bc750 Foreign function (null) ; WHICH SIGNAL ?
+ *  5: fp=0x7fa86423f630 pc=0x2713ce Foreign function (null)       ; verify_range
+ *  6: fp=0x7fa86423f6c0 pc=0x2703d3 Foreign function (null)       ; verify_heap
+ *  7: fp=0x7fa86423f780 pc=0x26e168 Foreign function collect_garbage
+ *  8: fp=0x7fa86423f7f0 pc=0x270298 Foreign function gc_and_save
+ *  9: fp=0x7fa86423f848 pc=0x52f93bf2 <??? type 45>::GC-AND-SAVE
+ * 10: fp=0x7fa86423f950 pc=0x52d32a7b <??? type 45>::SAVE-LISP-AND-DIE
+ * 11: fp=0x7fa86423fa00 pc=0x52a34179 <??? type 45>::SAVE-LISP-AND-DIE
+ *
+ * fatal error encountered in SBCL pid 9436 tid 9436:
+ * interrupt already pending
+ */
+
 static boolean
 can_handle_now(void *handler, struct interrupt_data *data,
                int signal, siginfo_t *info, os_context_t *context)
@@ -1339,7 +1358,8 @@ can_handle_now(void *handler, struct interrupt_data *data,
     struct thread *thread = get_sb_vm_thread();
 
     if (read_TLS(INTERRUPT_PENDING,thread) != NIL)
-        lose("interrupt already pending");
+        lose("interrupt already pending when sig%d received, pc=%p", signal,
+             (void*)os_context_pc(context));
     if (thread_interrupt_data(thread).pending_handler)
         lose("there is a pending handler already (PA)");
     if (data->gc_blocked_deferrables)
@@ -1359,7 +1379,7 @@ can_handle_now(void *handler, struct interrupt_data *data,
         event3("can_handle_now(%p,%d): deferred (RACE=%d)", handler, signal,
                in_leaving_without_gcing_race_p(thread));
         store_signal_data_for_later(data,handler,signal,info,context);
-        write_TLS(INTERRUPT_PENDING, T,thread);
+        write_TLS(INTERRUPT_PENDING, LISP_T, thread);
         answer = 0;
     }
     /* a slightly confusing test. arch_pseudo_atomic_atomic() doesn't
@@ -1409,11 +1429,11 @@ sig_stop_for_gc_handler(int __attribute__((unused)) signal,
      * pseudo atomic until gc is finally allowed. */
     if (read_TLS(GC_INHIBIT,thread) != NIL) {
         event0("stop_for_gc deferred for *GC-INHIBIT*");
-        write_TLS(STOP_FOR_GC_PENDING,T,thread);
+        write_TLS(STOP_FOR_GC_PENDING, LISP_T, thread);
         return;
     } else if (arch_pseudo_atomic_atomic(context)) {
         event0("stop_for_gc deferred for PA");
-        write_TLS(STOP_FOR_GC_PENDING,T,thread);
+        write_TLS(STOP_FOR_GC_PENDING, LISP_T, thread);
         arch_set_pseudo_atomic_interrupted(context);
         maybe_save_gc_mask_and_block_deferrables
             (os_context_sigmask_addr(context));
@@ -1542,7 +1562,6 @@ extern int *os_context_flags_addr(os_context_t *context);
 
 extern lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs);
 extern void post_signal_tramp(void);
-extern void call_into_lisp_tramp(void);
 
 void
 arrange_return_to_c_function(os_context_t *context,
@@ -1605,7 +1624,7 @@ arrange_return_to_c_function(os_context_t *context,
 #if defined(LISP_FEATURE_DARWIN)
     uint32_t *register_save_area = (uint32_t *)os_allocate(0x40);
 
-    /* 1. os_validate (malloc/mmap) register_save_block
+    /* 1. allocate (malloc/mmap) register_save_block
      * 2. copy register state into register_save_block
      * 3. put a pointer to register_save_block in a register in the context
      * 4. set the context's EIP to point to a trampoline which:
@@ -1731,13 +1750,7 @@ arrange_return_to_c_function(os_context_t *context,
 void
 arrange_return_to_lisp_function(os_context_t *context, lispobj function)
 {
-#if defined(LISP_FEATURE_DARWIN) && defined(LISP_FEATURE_X86)
-    arrange_return_to_c_function(context,
-                                 (call_into_lisp_lookalike)call_into_lisp_tramp,
-                                 function);
-#else
     arrange_return_to_c_function(context, call_into_lisp, function);
-#endif
 }
 
 // These have undefined_alien_function tramp in x-assem.S
@@ -1755,6 +1768,7 @@ undefined_alien_function(void)
 }
 #endif
 
+#ifndef LISP_FEATURE_WIN32
 void lower_thread_control_stack_guard_page(struct thread *th)
 {
     protect_control_stack_guard_page(0, th);
@@ -1771,12 +1785,14 @@ void reset_thread_control_stack_guard_page(struct thread *th)
     th->state_word.control_stack_guard_page_protected = 1;
     fprintf(stderr, "INFO: Control stack guard page reprotected\n");
 }
+#endif
 
 boolean
 handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
 {
     struct thread *th=get_sb_vm_thread();
 
+#ifndef LISP_FEATURE_WIN32
     if(addr >= CONTROL_STACK_HARD_GUARD_PAGE(th) &&
        addr < CONTROL_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
         fake_foreign_function_call(context);
@@ -1827,8 +1843,10 @@ handle_guard_page_triggered(os_context_t *context,os_vm_address_t addr)
         reset_thread_control_stack_guard_page(th);
         return 1;
     }
-    else if(addr >= BINDING_STACK_HARD_GUARD_PAGE(th) &&
-            addr < BINDING_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
+    else
+#endif // !LISP_FEATURE_WIN32
+    if(addr >= BINDING_STACK_HARD_GUARD_PAGE(th) &&
+       addr < BINDING_STACK_HARD_GUARD_PAGE(th) + os_vm_page_size) {
         lose("Binding stack exhausted");
     }
     else if(addr >= BINDING_STACK_GUARD_PAGE(th) &&
@@ -2026,7 +2044,6 @@ interrupt_init(void)
     ll_install_handler(SIGINFO, sigdump_eventlog);
 #endif
     int __attribute__((unused)) i;
-    SHOW("entering interrupt_init()");
     sigemptyset(&deferrable_sigset);
     sigemptyset(&blockable_sigset);
     sigemptyset(&gc_sigset);
@@ -2066,6 +2083,10 @@ lisp_memory_fault_error(os_context_t *context, os_vm_address_t addr)
     /* If we lose on corruption, provide LDB with debugging information. */
     fake_foreign_function_call(context);
 
+    /* If it's a store to read-only space, it's not "corruption", so don't say that.
+     * Lisp will change its wording of the memory-fault-error string */
+
+    if (!readonly_space_p((uword_t)addr)) {
     /* To allow debugging memory faults in signal handlers and such. */
 #ifdef ARCH_HAS_STACK_POINTER
     char* pc = (char*)os_context_pc(context);
@@ -2086,6 +2107,7 @@ lisp_memory_fault_error(os_context_t *context, os_vm_address_t addr)
     corruption_warning_and_maybe_lose("Memory fault at %p (pc=%p)",
                                       addr, (void*)os_context_pc(context));
 #endif
+    }
 
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
     /* Holy hell is this more obfuscated than necessary when using
